//
// C++ Implementation: cavity2
//
// Description:
//
//
// Author: mirco <Mirco.till@asamnet.de>, (C) 2008
//
// Copyright: See COPYING file that comes with this distribution
//
//
#include "cavity2.h"
/** Initialize a cavity based ona Monte Carlo run and graph search

Initializes a cavity. Takes two grids containig information about the cavity belongings of each point (a result of the graph search) and a grid containing the number of MC points on each grid point (a result of an MC run) as well as the overall number of MC steps. Combining these information all grid points of the cavity are found and the volume is calculated based on the MC run.
\param cavgrid Containing the informaiton about the points belonging to cavities
\param solvgird Containing the information about the number of MC steps mapped on each point
\param ncav Number of cavity which should be extracted from cavgrid
\param nmc Overall number of MC steps the solvgrid is based on
\param occ Occupancy or charge for the dummy atoms
\param dummy_rad Radius for the dummy atoms
\param k Kind of the new cavity. 1 for cavity, 2 for cleft
\return void
*/
void cavity2::init(iGrid &cavgrid, iGrid &solvgrid, int ncav, int nmc, int occ, double dummy_rad, int k) {
  cav_point *tmp_point;
  double local_volume = 0;
  running_number = occ;
  kind = k;
  box.min(1) = cavgrid.imin[1];
  box.min(2) = cavgrid.imin[2];
  box.min(3) = cavgrid.imin[3];
  box.max(1) = cavgrid.imax[1];
  box.max(2) = cavgrid.imax[2];
  box.max(3) = cavgrid.imax[3];
  box.spacing = cavgrid.spacing;
  cout << box.spacing << " " << solvgrid.spacing << endl;
  double boxvol = (box.max(1) - box.min(1)) * (box.max(2) - box.min(2)) * (box.max(3) - box.min(3));
  for (int i = cavgrid.imin[1]; i <= cavgrid.imax[1]; i++) {
    for (int j = cavgrid.imin[2]; j <= cavgrid.imax[2]; j++) {
      for (int k = cavgrid.imin[3]; k <= cavgrid.imax[3]; k++) {
        if (cavgrid.grid[i][j][k] == ncav) {
          local_volume += solvgrid.grid[i][j][k];
          tmp_point = new cav_point(i, j, k, occ, dummy_rad);
          points.add_tail(tmp_point);
        };
      }
    };
  }
  volume = boxvol * (local_volume / nmc);
  check_real_box();
}
void cavity2::print_points_min_max() {
  int minx,miny,minz,maxx,maxy,maxz;
  minx = miny = minz = (int) 1E5;
  maxx = maxy = maxz = (int) -1E5;
  adv_doubly_LL<cav_point>::iterator iter = points.begin();
  while (iter != NULL) {
  if((*iter)->int_x > maxx){maxx = (*iter)->int_x;}
  if((*iter)->int_y > maxy){maxy = (*iter)->int_y;}
  if((*iter)->int_z > maxz){maxz = (*iter)->int_z;}
  if((*iter)->int_x < minx){minx = (*iter)->int_x;}
  if((*iter)->int_y < miny){miny = (*iter)->int_y;}
  if((*iter)->int_z < minz){minz = (*iter)->int_z;}

  iter++;
  }
  cout << "points max: " << maxx << " " << maxy << " " << maxz << endl;
  cout << "points min: " << minx << " " << miny << " " << minz << endl;
}
/** Check, if the cavity contains point at the box border

This check is necessary, since the box is choosen one point larger than the actual cavity. If the cavity contains points on the border of the box, it is most likely an artifact of another cavity, close enough to also take space inside our box bot not belonging to our cavity at all.
\return int 0 if cavity points are found on the box border, 1 otherwise
*/
int cavity2::check_border_points() {
  adv_doubly_LL<cav_point>::iterator iter = points.begin();
  //cout << "cav: " << running_number << endl;
  //print_box();
  while (iter != NULL) {
    if ((*iter)->int_x == box.min(1) || (*iter)->int_x == box.max(1)) {
      //cout << "Touches xbox" << endl;
      return 0;
    }
    if ((*iter)->int_y == box.min(2) || (*iter)->int_y == box.max(2)) {
      //cout << "Touches ybox" << endl;
      return 0;
    }
    if ((*iter)->int_z == box.min(3) || (*iter)->int_z == box.max(3)) {
      //cout << "Touches zbox" << endl;
      return 0;
    }
    iter++;
  }
  return 1;
}

/** redifining the cavity box

the box is choosen on point larger on each site by the MC algorithm. Therefore this function checks the real dimensions of a surrounding box. and saves the result in the real_box class.
\return void
*/
void cavity2::check_real_box() {
  real_box.min(1) = box.max(1);
  real_box.min(2) = box.max(2);
  real_box.min(3) = box.max(3);
  real_box.max(1) = box.min(1);
  real_box.max(2) = box.min(2);
  real_box.max(3) = box.min(3);
  adv_doubly_LL<cav_point>::iterator iter = points.begin();
  while (iter != NULL) {
    if ((*iter)->int_x > real_box.max(1)){real_box.max(1) = (*iter)->int_x;}
    if ((*iter)->int_x < real_box.min(1)){real_box.min(1) = (*iter)->int_x;}
    if ((*iter)->int_y > real_box.max(2)){real_box.max(2) = (*iter)->int_y;}
    if ((*iter)->int_y < real_box.min(2)){real_box.min(2) = (*iter)->int_y;}
    if ((*iter)->int_z > real_box.max(3)){real_box.max(3) = (*iter)->int_z;}
    if ((*iter)->int_z < real_box.min(3)){real_box.min(3) = (*iter)->int_z;}
    iter++;
  }
  real_box.spacing = box.spacing;
}

/** Print information about a grid

\param b the grid we want to print some information fork(
*/
void cavity2::print_box(grid &b) {
  cout << "box dimensions" << endl;
  cout << "x: " << b.min(1) << " " << b.max(1) << endl;
  cout << "y: " << b.min(2) << " " << b.max(2) << endl;
  cout << "z: " << b.min(3) << " " << b.max(3) << endl;
  cout << "box.spacing " << b.spacing << endl;
}

/** Check if the cavity contains a point

\param p The point to check
\param ratio The ratio between both cavities spacing, if necessary
\return int 1 if the point is found inside the cavity, 0 otherwise
*/
int cavity2::contains_point(cav_point &p, double ratio) {
  adv_doubly_LL<cav_point>::iterator iter = points.begin();
  while (iter != NULL) {
    if ((*iter)->compare(p, ratio)) {
      return 1;
    }
    iter++;
  }
  return 0;
}

/** Check if a cavity contains also points from another cavity.

Pretty time consuming for large cavities, this function checks each point of the cavity with each point of a second cavity if the twopoints are the same. To reduce the computation time, it first of all checks the boxes. if they don't overlapp, the cavities can't. If the boxes are overlapping each point is checked.
\param other The second cavity
\return 1 If the cavities are overlapping, 0 otherwise
*/
int cavity2::check_duplicate(cavity2 &other) {
  adv_doubly_LL<cav_point>::iterator point_it = points.begin();
  double ratio = other.box.spacing / box.spacing;
  // Check overlapping boxes
  if (check_overlapping_boxes(box, other.box)) {
    while (point_it != NULL) {
      if (other.contains_point(**point_it, ratio)) {
        return 1;
      }
      point_it++;
    }
    return 0;
  } else {
    return 0;
  }
}

/** Merge two cavities

If we found two overlapping cavities (see check_duplicate()), it's very instructive to merge these cavities. for each point of this cavity, we check, if the second cavity got this point. If not, we add the point to the other cavity. This cavity should be deleted afterwards.
\param other Second cavity and resulting cavity
*/
void cavity2::merge_with(cavity2 &other) {
  adv_doubly_LL<cav_point>::iterator point_it = points.begin();
  double ratio = other.box.spacing / box.spacing;
  while (point_it != NULL) {
    if (!other.contains_point(**point_it, ratio)) {
      cav_point *tmp = new cav_point((*point_it)->x, (*point_it)->y, (*point_it)->z);
      other.points.add_tail(tmp);
    }
    point_it++;
  }
  if (box.min(1) < other.box.min(1)){other.box.min(1) = box.min(1);}
  if (box.min(2) < other.box.min(2)){other.box.min(2) = box.min(2);}
  if (box.min(3) < other.box.min(3)){other.box.min(3) = box.min(3);}
  if (box.max(1) > other.box.max(1)){other.box.max(1) = box.max(1);}
  if (box.max(2) > other.box.max(2)){other.box.max(2) = box.max(2);}
  if (box.max(3) > other.box.max(3)){other.box.max(3) = box.max(3);}
  other.check_real_box();
}

/** Printing only a slice of a whole cavity.

If we have the whole solvent in one cavity, we can print the desired  membrane slice with this function. Two files are generated, one containing the membrane slice with no point above the maximum z value and no point below the minimum z value. All points filtered out while printing the first file (the membrane slice) are printed to the second file as membrane cutof
\param out The file for the membrane slice
\param alt_out The file for the membrane cutof
\param z_start The minimum z coordinate
\param z_stop The maximum z coordinate
\param start_res Starting value for the residue counter
\return void
*/
void cavity2::print_pqr_membrane_slice(ostream &out, ostream &alt_out, double z_start, double z_stop, int start_res) {
  int natoms = 1;
  int natoms_alt = 1;
  int start_res_alt = start_res;
  adv_doubly_LL<cav_point>::iterator iter = points.begin();
  while (iter != NULL) {
    if ((*iter)->z*box.spacing >= z_start && (*iter)->z*box.spacing <= z_stop) {
      write_line_pqr(out, "ATOM", natoms, "DUM", "DUM", start_res, (*iter)->x*box.spacing, (*iter)->y*box.spacing, (*iter)->z*box.spacing, (*iter)->charge, (*iter)->rad);
      if (natoms % 100 == 0){start_res++;}
      natoms++;
    } else {
      write_line_pqr(alt_out, "ATOM", natoms_alt, "DUM", "DUM", start_res_alt, (*iter)->x*box.spacing, (*iter)->y*box.spacing, (*iter)->z*box.spacing, (*iter)->charge, (*iter)->rad);
      if (natoms_alt % 100 == 0){start_res_alt++;}
      natoms_alt++;
    }
    iter++;
  }
}

/** Print cavity as pqr

Simply print all cavity points in pqr format
\param out The output file
\param start_res The residue to start with
\return The last redsidue counter
*/
int cavity2::print_pqr(ostream &out, int start_res) {
  int natoms = 1;
  adv_doubly_LL<cav_point>::iterator iter = points.begin();
  while (iter != NULL) {
    write_line_pqr(out, "ATOM", natoms, "DUM", "DUM", start_res, (*iter)->x*box.spacing, (*iter)->y*box.spacing, (*iter)->z*box.spacing, (*iter)->charge, (*iter)->rad);
    natoms++;
    iter++;
  }
  return start_res;
}

/** Print water molecules in this cavity

Printing all the water molecules O-atoms in pqr format
\param out The output file
\param start_res Starting residue counter
\return residue counter at the end of the day
*/
int cavity2::print_waters(ostream &out, int start_res) {
  int natoms = 1;
  int nres = start_res;
  adv_doubly_LL<cav_point>::iterator iter = waters.begin();
  while (iter != NULL) {
    out << "ATOM" << setw(7) << nres << setw(10) << " O   HOH" << setw(6) << nres;
    out << setw(10) << fixed << setprecision(3) << (*iter)->x*box.spacing;
    out << setw(8) << fixed << setprecision(3) << (*iter)->y*box.spacing;
    out << setw(8) << fixed << setprecision(3) << (*iter)->z*box.spacing;
    out << setw(8) << fixed << setprecision(3) << 0.0;
    out << setw(8) << fixed << setprecision(3) << 0.0;
    out << endl;
    natoms++;
    nres++;
    iter++;
  }
  return nres;
}
/** Placing water molecules in the cavity volume.

Water molecules oxygen postions are paced randomly in the cavity volume. Afterwards, the positions are relaxed by a MC algorithm. This algorithms takes the Sum of the distances between the all pairs of water molecules and the distances of all water molecules to the borders of the box as a function. This function is maximised by the Monte Carlo algorithm. The aim is to place the water molecules equally distributed across the membrane volume.
\param water vol Defined volume of one water molecule.
\return int Number of placed water molecules.
*/

int cavity2::add_water_molecules2(double water_vol) {
  tensor<char> water_grid;// map space occupied by the cavity
  water_grid.init_arr(box.min(1), box.max(1), box.min(2), box.max(2), box.min(3), box.max(3), 0);
  adv_doubly_LL<cav_point>::iterator iter = points.begin();
  while (iter != NULL) {
    water_grid((*iter)->int_x, (*iter)->int_y, (*iter)->int_z) = 1;
    iter++;
  }
  // Define how many water molecules we have to place
  int nwaters = (int) round(volume / water_vol);
  //cout << "searching for " << nwaters << " in " << volume << endl;
  matrix<int> placed_waters(1, nwaters, 1, 3, 0); // coords of placed water molecules
  time_t tim;
  long *idum, id;
  id = time(&tim) * -1;
  idum = &id;
  int xlength = box.max(1) - box.min(1);
  int ylength = box.max(2) - box.min(2);
  int zlength = box.max(3) - box.min(3);
  int nsteps;
  // Plcae the water molecules randomly in the cavity
  int n = 0;
  while (n <= nwaters) {
    int flag = 0;
    nsteps = 0;
    while (flag == 0) {
      nsteps++;
      int x = (int)(box.min(1) + ran2(idum) * xlength);
      int y = (int)(box.min(2) + ran2(idum) * ylength);
      int z = (int)(box.min(3) + ran2(idum) * zlength);
      /*
      cout << "check " << x << " " << y << " " << z << endl;
      cout << "len " << xlength << " " << ylength << " " << zlength << endl;
      cout << "box " << box.min(1) << " " << box.max(1)<< " "  << box.min(2) << " " << box.max(2) << " " << box.min(3) << " " << box.max(3) <<endl;
      */
      // Is the point a cavity point??
      if (water_grid(x, y, z) == 1) {
        n++;
        placed_waters(n, 1) = x;
        placed_waters(n, 2) = y;
        placed_waters(n, 3) = z;
        flag = 1;

      };
      if (n >= nwaters){break;}
    }
    if (n >= nwaters){break;}
  };

  // Now releax them a little bit
  int flag = 0;
  int to_change, coor, new_val[4];
  double sum = -1, last_sum = -1, pm_one;
  vec<int> count(1, nwaters, 0);
  while (flag < 10000 && nwaters > 1) {
    //change one position randomly
    to_change = (int)(1 + (ran2(idum) * (nwaters)));
    count(to_change) += 1;
    coor = (int)(1 + (ran2(idum) * 2));
    pm_one = ran2(idum);
    if (pm_one > 0.5) {
      pm_one = 1;
    } else {
      pm_one = -1;
    }
    new_val[1] = placed_waters(to_change, 1);
    new_val[2] = placed_waters(to_change, 2);
    new_val[3] = placed_waters(to_change, 3);
    new_val[coor] = placed_waters(to_change, coor) + (int) pm_one;
    // Point is valid
    if (new_val[coor] > box.min(coor) && new_val[coor] < box.max(coor) && water_grid(new_val[1], new_val[2], new_val[3]) == 1) {
      placed_waters(to_change, coor) += (int) pm_one;
      sum = 0;
      //cout << "accepted " << pm_one << " " << to_change << " " << sum <<  endl;
      for (int i = 1; i <= nwaters; i++) {
        //distances to the box borders
        sum += get_dist(placed_waters(i, 1), placed_waters(i, 2), placed_waters(i, 3), box.min(1), placed_waters(i, 2), placed_waters(i, 3));
        sum += get_dist(placed_waters(i, 1), placed_waters(i, 2), placed_waters(i, 3), box.max(1), placed_waters(i, 2), placed_waters(i, 3));
        sum += get_dist(placed_waters(i, 1), placed_waters(i, 2), placed_waters(i, 3), placed_waters(i, 1), box.min(2), placed_waters(i, 3));
        sum += get_dist(placed_waters(i, 1), placed_waters(i, 2), placed_waters(i, 3), placed_waters(i, 1), box.max(2), placed_waters(i, 3));
        sum += get_dist(placed_waters(i, 1), placed_waters(i, 2), placed_waters(i, 3), placed_waters(i, 1), placed_waters(i, 2), box.min(3));
        sum += get_dist(placed_waters(i, 1), placed_waters(i, 2), placed_waters(i, 3), placed_waters(i, 1), placed_waters(i, 2), box.max(3));
        //distances to all other water molecules in this cav
        for (int j = 1; j <= nwaters; j++) {
          sum += get_dist(placed_waters(i, 1), placed_waters(i, 2), placed_waters(i, 3), placed_waters(j, 1), placed_waters(j, 2), placed_waters(j, 3));
        };
      };

      if (sum > last_sum) {
        // keep changes
        last_sum = sum;
      } else {
        // discard changes
        placed_waters(to_change, coor) -= (int) pm_one;
      }
      flag++;
    };
  };
  for (int n = 1; n <= nwaters; n++) {
    waters.add_tail(new cav_point(placed_waters(n, 1), placed_waters(n, 2), placed_waters(n, 3)));
  }
  //cout << count << endl;
  return nwaters;
}

/** Making a layer from a box

Since the solvent cavity is taken from the gird it is shaped as a box. Now we reduce pr enlarge this box to form a layer around the protein.
\param coors. Atom coordinates of the protein
\param natoms Number of atoms
\param range Thickness of the layer to create
\return void
*/
void cavity2::reduce_cav_to_layer(double **coors, int natoms, double range) {
  adv_doubly_LL<cav_point>::iterator iter = points.begin();
  int flag;
  double ndeleted = 0;
  double nkept = 0;
  range = range * range;
  // For all cavity points
  while (iter != NULL) {
    flag = 0;
    // check the distance to the atoms
    for (int i = 1; i <= natoms; i++) {
      if (get_sqr_dist((*iter)->x*box.spacing, (*iter)->y*box.spacing, (*iter)->z*box.spacing, coors[i][1], coors[i][2], coors[i][3]) < range) {
        flag = 1;
        break;
      }
    }
    /*
    if(flag == 0){
    to_delete = iter;
    iter++;
    ndeleted++;
    points.delete_element(to_delete);
      }else{
    nkept++;
    iter++;
      }
      */
    // If there is no atom in the desired range, delete this point
    if (flag == 0) {
      ndeleted++;
      points.delete_element(iter);
    } else {
      nkept++;
      iter++ ;
    }

  }
  cout << ndeleted << " dummy atoms deleted, " << nkept << " dummy atoms kept. This means a reduction by " << fixed << setprecision(2) << (ndeleted / (ndeleted + nkept))*100 << "%" << endl;
}

/** Take the solvent box and extrace a slice with defined z coordinates

Works only, if the protein is oriented correctly in z directino. Takes the whole solvent box and extracts a slice from this box. The slice contains all points between the minimal and the maximal z coordinate. Afterwards the core region is checked. For a defined headgroup regin, each point is evaluated if its close enough to a core region. This maybe reduce the problem of a membrane floating into a channel, if the membrane can float over the channel borders. Finally we run a cavity search again, since we have maybe cut of some regions from the membrane, for example the overflooded region occupying our channel
\param start_cav The original cavity
\param zmin Minimum z coordinate of the membrane slice to extract
\param zmax Maximum z coordinate of the membrane slice to extract
\param dummy_rad Dummy radius for the membrane atoms
\param core_min Minimum z coordinate of the copre region
\param core_max Maximum z coordinate of the copre region
\param core_dim Maximal allowed distance of the headgroup atoms the to core region
\return void
*/
void cavity2::extract_membrane_slice(cavity2 &start_cav, double zmin, double zmax, double dummy_rad, double core_min, double core_max, double core_dim) {
  // define new box dimensions
  running_number = start_cav.running_number;
  box.min(1) = start_cav.box.min(1);
  box.min(2) = start_cav.box.min(2);
  box.min(3) = (int) zmin - 1;
  box.max(1) = start_cav.box.max(1);
  box.max(2) = start_cav.box.max(2);
  box.max(3) = (int) zmax + 1;
  box.spacing = start_cav.box.spacing;
  check_real_box();
  adv_doubly_LL<cav_point>::iterator iter = start_cav.points.begin();
  // All points
  while (iter != NULL) {
    // If they between zmin und zmax, add them to the slice
    if ((*iter)->z*box.spacing > zmin && (*iter)->z*box.spacing < zmax) {
      cav_point *tmp_point = new cav_point((*iter)->x, (*iter)->y, (*iter)->z, start_cav.running_number, dummy_rad);
      points.add_tail(tmp_point);
      start_cav.points.delete_element(iter);
    } else {
      iter++;
    }
  }
  //For each membrane point, check if its close enough to the core region.
  if (core_min > zmin && core_max < zmax && core_dim > 0) {
    cout << "Checking core region" << endl;
    iter = points.begin();
    int flag;
    double sqr_core_dist = core_dim * core_dim;
    // For all points of the slice
    while (iter != NULL) {
      flag = 0;
      // if this point is inside the core region, means smaller than core_min or larger than core_max
      if ((*iter)->z < core_min || (*iter)->z > core_max) {
        adv_doubly_LL<cav_point>::iterator iter2 = points.begin();
        // For all points of the slice  which are inside the core region
        while (iter2 != NULL) {
          if ((*iter2)->z > core_min && (*iter2)->z < core_max) {
            if (get_sqr_dist((*iter)->x*box.spacing, (*iter)->y*box.spacing, (*iter)->z*box.spacing, (*iter2)->x*box.spacing, (*iter2)->y*box.spacing, (*iter2)->z*box.spacing) < sqr_core_dist) {
              flag = 1;
              break;
            }
          };
          if (flag == 1){break;}
          iter2++;
        }
        // If there is no point in the core region close enough, put this point back to the start_cav
        if (flag == 0) {
          cav_point *tmp_point = new cav_point((*iter)->x, (*iter)->y, (*iter)->z, start_cav.running_number, dummy_rad);
          start_cav.points.add_tail(tmp_point);
          points.delete_element(iter);
        } else {
          iter++;
        }
      } else{iter++;}
    };
  }

  // Initialise the grids for the graph search
  iGrid memb_solvgrid, memb_cavgrid;
  set_iGrid(memb_cavgrid, box.min(1), box.min(2), box.min(3), box.max(1), box.max(2), box.max(3));
  memb_cavgrid.grid = i3tensor(box.min(1), box.max(1), box.min(2), box.max(2), box.min(3), box.max(3));
  memb_cavgrid.spacing = box.spacing;
  initialize_iGrid(memb_cavgrid, 0);

  set_iGrid(memb_solvgrid, box.min(1), box.min(2), box.min(3), box.max(1), box.max(2), box.max(3));
  memb_solvgrid.grid = i3tensor(box.min(1), box.max(1), box.min(2), box.max(2), box.min(3), box.max(3));
  memb_solvgrid.spacing = box.spacing;
  initialize_iGrid(memb_solvgrid, 0);
  iter = points.begin();
  cout << "slicing membrane" << endl;
  // Chek, if we have more than one graph in the membrane region
  while (iter != NULL) {
    memb_solvgrid.grid[(*iter)->int_x][(*iter)->int_y][(*iter)->int_z] = 1;
    iter++;
  };
  cout << search_cav(memb_solvgrid, memb_cavgrid) << " clusters found during membrane buildup" << endl;
  iter = points.begin();
  while (iter != NULL) {
    // Only the first graph found contains the membrane, all other graphs are artifacts and also put back to the start_cav
    if (memb_cavgrid.grid[(*iter)->int_x][(*iter)->int_y][(*iter)->int_z] > 1) {
      cav_point *tmp_point = new cav_point((*iter)->x, (*iter)->y, (*iter)->z, start_cav.running_number, dummy_rad);
      start_cav.points.add_tail(tmp_point);
      points.delete_element(iter);
    } else {
      iter++;
    };
  };

}

/** Simple output for a point
*/
ostream &operator<<(ostream &output, cav_point &c) {
  output << c.x << " " << c.y << " " << c.z << endl;
  return output;
}

/** Increase the cavity as well as all points to a given box dimension

Afterwards we can reduce the cavity to a layer and its sure, that there are at least enough points on the narrowest point
\param xmin Minimum coor of the new box
\param xmax Maximum coor of the new box
\param ymin Minimum coor of the new box
\param ymax Maximum coor of the new box
\param zmin Minimum coor of the new box
\param zmax Maximum coor of the new box
\param rad Dummz atom radius
\return void
*/
void cavity2::increase_cav_to_dimension(int xmin, int xmax, int ymin, int ymax, int zmin, int zmax, double rad) {
  for (int x = xmin; x <= xmax; x++) {
    for (int y = ymin; y <= ymax; y++) {
      for (int z = zmin; z <= zmax; z++) {
        // If this point is not in the cavity...
        if (x < real_box.min(1) || x >= real_box.max(1) || y < real_box.min(2) || y >= real_box.max(2)) {
          points.add_tail(new cav_point(x, y, z, 0, rad));
        }
      }
    }
  }
  // set_new_boxes
  check_real_box();
  box = real_box;
}

/** Check if two boxes are overlapping */
int check_overlapping_boxes(grid &b1, grid &b2) {
  if (b1.min(1) > b2.max(1) || b2.min(1) > b1.max(1)){return 0;}
  if (b1.min(2) > b2.max(2) || b2.min(2) > b1.max(2)){return 0;}
  if (b1.min(3) > b2.max(3) || b2.min(3) > b1.max(3)){return 0;}
  return 1;
}

