//
// C++ Interface: cavity2
//
// Description:
//
//
// Author: mirco <Mirco.till@asamnet.de>, (C) 2008
//
// Copyright: See COPYING file that comes with this distribution
//
//
#ifndef CAVITY2_H
#define CAVITY2_H
#include "util.h"
#include "util_c.h"
#include "nrutil.h"
#include "grid.h"
#include "random.h"
#include "grid.h"
#include "grid_helper.h"
#include "cavity_search.h"

/** One (Grid-) Point of a cavity

Representing a grid point of a cavity, xyz coordinates, charge and radius
*/
class cav_point {
  public:
    double x; /**< x-coordinate */
    double y; /**< y-coordinate */
    double z; /**< z-coordinate */
    double charge; /**< charge */
    double rad; /**< radius */
    int int_x; /**< x-coordinate rounded to integer*/
    int int_y; /**< y-coordinate rounded to integer*/
    int int_z; /**< z-coordinate rounded to integer*/
    /** Constructor

    Initialising all data to 0
    */
    cav_point(): x(0), y(0), z(0), charge(0), rad(0), int_x(0), int_y(0), int_z(0) {};
    /** Constructor
    Initialising all data to 0
    \param x1 initial x coordinate
    \param y1 initial y coordinate
    \param z1 initial z coordinate
    \param c initial charge default 0
    \param r initial radius default 0
     */
    cav_point(double x1, double y1, double z1, double c = 0, double r = 0): x(x1), y(y1), z(z1), charge(c), rad(r), int_x((int) round(x)), int_y((int) round(y)), int_z( (int) round(z)) {};
    /** Comparison

      Compares xyz coordinates between this point and another point.
    \return bool
    */
    bool operator==(const cav_point &other) {
      return (int_x == other.int_x && int_y == other.int_y && int_z == other.int_z);
    }
    /** Comparison

    Compares xyz coordinates between this point and another point weighted by a given ratio
    \param other Second point to evaluate
    \param ratio weighting for the points before comparing to the other point
    \return bool
     */
    bool compare(cav_point &other, double ratio = 1) {
      int tmp_x = (int) round(x * ratio);
      int tmp_y = (int) round(y * ratio);
      int tmp_z = (int) round(z * ratio);
      return (tmp_x == other.int_x && tmp_y == other.int_y && tmp_z == other.int_z);
    };

    friend ostream &operator<<(ostream &output, cav_point &c);
};

/** Defining the variables for a simple grid

  There are no grid points in this class, only the surrounding variables for a defined grid
*/
class grid {
  public:
    vec<int> min; /**< Minimum coordinates of the gird min(1) = minimum x value */
    vec<int> max; /**< Maximum coordinates of the gird max(1) = maximum x value */
    float spacing; /**< Grid Spacing */
    /** Simple constructor

    Initialise everything to 0
    */
    grid(): min(1, 3, 0), max(1, 3, 0), spacing(0) {};
    /** Assignment Operator

    Assigning min, max and spacing to another grid
    */
    grid &operator=(const grid &g) {
      for (int i = 1; i <= 3;i++) {
        min(i) = g.min(i);
        max(i) = g.max(i);
        spacing = g.spacing;
      }
      return *this;
    }
    /** Copy Constructor

    Constructing a grid based on another girds data
     */
    grid(const grid &g) {
      for (int i = 1; i <= 3;i++) {
        min(i) = g.min(i);
        max(i) = g.max(i);
        spacing = g.spacing;
      }
    }
    /** Desturctor

    Free memory of min and max vectors
    */
    ~grid() {
      min.kill_arr();
      max.kill_arr();
    }
};

/** Definition of a single cavity

Holding all data available for a given cavity. Calculates the volume based on a previous Monte Carlo run. Also keeping the coordinates of the water molecules of this cavity
*/
class cavity2 {
  private:
    double volume; /**< Volume of this cavity derived by MC */
    int running_number; /**< Running number of this cavity in a list */
    int npoints; /**< Number of cavity points = length of points list */
    grid box;/**< Information about the surrounding box derived from Monte Carlo vars*/
    char kind; /**< 1 is cavity, 2 is cleft */
    grid real_box; /**< Information about the surrounding box derived from all points (compare to MC derived box) */
    adv_doubly_LL<cav_point> points; /**< List of all points */
    adv_doubly_LL<cav_point> waters; /**< List of all water molecules */
  public:
    void init(iGrid &cavgrid, iGrid &solvgrid, int ncav, int nmc, int occ, double dummy_rad, int k);
    int print_pqr(ostream &out, int start_res = 1);
    void print_pqr_membrane_slice(ostream &out, ostream &alt_out, double z_start, double z_stop, int start_res = 1);
    int print_waters(ostream &out, int start_res = 1);
    /** Return cavity volume
    \return cavity volume as double
    */
    double get_volume(){return volume;}
    /** Return number of points
    \return number of points as int
     */
    int get_npoints(){return npoints;}
    /** Return running number
    \return running number as int
     */
    int get_running_num(){return running_number;}
    int contains_point(cav_point &p, double ratio = 1);
    int check_duplicate(cavity2 &other);
    int add_water_molecules2(double water_vol);
    void reduce_cav_to_layer(double **coors, int natoms, double range);
    void extract_membrane_slice(cavity2 &start_cav, double zmin, double zmax, double dummy_rad, double core_min, double core_max, double core_dim);
    void increase_cav_to_dimension(int xmin, int xmax, int ymin, int ymax, int zmin, int zmax, double rad);
    void check_real_box();
    int check_border_points();
    void merge_with(cavity2 &other);
    /** Print this box
    \return void
    */
    void print_box(){print_box(box);}
    void print_box(grid &b);
    void print_points_min_max();
    /** Print this real box
    \return void
    */
    void print_real_box(){print_box(real_box);}
    /** Return minimum coor of the box
    \param n 1=x, 2=y, 3=z
    */
    int get_min(int n){return box.min(n);};
    /** Return maximum coor of the box
    \param n 1=x, 2=y, 3=z
     */
    int get_max(int n){return box.max(n);};
    /** Return cavity kind
    \param kind as char, 1=cavity,2=cleft
     */
    char get_kind(){return kind;}
    /** Destructor

    free linked lists
    */
    ~cavity2() {
      points.clear();
      waters.clear();
    }
};
int check_overlapping_boxes(grid &b1, grid &b2);
#endif
