/* Matthias Ullmann 					*/
/* Berlin, 04.10.1995					*/
/* 							*/
/* Cell base neighbor list for Monte Carlo Simulations	*/
/* 							*/
/* Parameter suggestion:				*/
/* for cut on 10.0 and cutoff 12.0 a grid spacing of 	*/
/* 4.0 has been found to be ideal			*/
/* .....................................................*/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include "nb-list.h"
#include "util_c.h"

#define NR_END 1
/* .....................................................*/
/* prints infos about the grid                          */
void iListGrid_info(FILE *fpout, iListGrid *grid)
{
fprintf(fpout,"MIN: "); PRINT_iVEC(fpout,(*grid).imin);
fprintf(fpout,"MAX: "); PRINT_iVEC(fpout,(*grid).imax);
fprintf(fpout,"Spacing  : %8.3f\n", (*grid).spacing);
fprintf(fpout,"Inverse Spacing  : %8.3f\n", (*grid).inv_spacing);
}

/* .....................................................*/
/* Allocating a vector of iList with 			*/
/* subscript range v[nl..nh] 				*/
iList *iList_vector(long nl, long nh)
{
	iList *v;

	v=(iList *)malloc((size_t) ((nh-nl+1+NR_END)*sizeof(iList)));
	if (!v) muerror("allocation failure in iList_vector()");
	return v-nl+NR_END;
}
/* .....................................................*/
/* Deallocating a vector of iList with 			*/
/* subscript range v[nl..nh] 				*/
void free_iList_vector(iList *v, long nl, long nh)
{
if (v!=NULL)    free((FREE_ARG) (v+nl-NR_END));
}
/* .....................................................*/
/* Allocating a 3D tensor of iList			*/
/* with subscript range v[nl..nh][ncl..nch][ndl..ndh]	*/
iList ***iList_3tensor(long nrl, long nrh, long ncl, long nch, long ndl, long ndh)
{
long  i,j,nrow=nrh-nrl+1,ncol=nch-ncl+1,ndep=ndh-ndl+1;
iList  ***t;

	/* allocate pointers to pointers to rows */
	t=(iList ***) malloc((size_t)((nrow+NR_END)*sizeof(iList**)));
	if (!t) muerror("allocation failure 1 in iList_3tensor()");
	t += NR_END;
	t -= nrl;

	/* allocate pointers to rows and set pointers to them */
	t[nrl]=(iList **) malloc((size_t)((nrow*ncol+NR_END)*sizeof(iList*)));
	if (!t[nrl]) muerror("allocation failure 2 in iList_3tensor()");
	t[nrl] += NR_END;
	t[nrl] -= ncl;

	/* allocate rows and set pointers to them */
	
	t[nrl][ncl]=(iList *) malloc((size_t)((nrow*ncol*ndep+NR_END)*sizeof(iList)));
	if (!t[nrl][ncl]) muerror("allocation failure 3 in iList_3tensor()");
	t[nrl][ncl] += NR_END;
	t[nrl][ncl] -= ndl;

	for(j=ncl+1;j<=nch;j++) t[nrl][j]=t[nrl][j-1]+ndep;
	for(i=nrl+1;i<=nrh;i++) {
		t[i]=t[i-1]+ncol;
		t[i][ncl]=t[i-1][ncl]+ncol*ndep;
		for(j=ncl+1;j<=nch;j++) t[i][j]=t[i][j-1]+ndep;
	}

	/* return pointer to array of pointers to rows */
	return t;
}

/* .....................................................*/
/* Deallocating a 3D tensor of iList			*/
/* with subscript range v[nl..nh][ncl..nch][ndl..ndh]	*/
void free_iList_3tensor(iList ***t, long nrl, long nrh, long ncl, 
                        long nch, long ndl, long ndh)
{
if (t!=NULL)    
        {free((FREE_ARG) (t[nrl][ncl]+ndl-NR_END));
         free((FREE_ARG) (t[nrl]+ncl-NR_END));
         free((FREE_ARG) (t+nrl-NR_END));
        }
}
/* .....................................................*/
/* Initialzes a vector of iList of the 	length 'length'	*/
/* Initialization: *next =NULL				*/
/*		   *prev =NULL				*/
/*		   number=counting			*/
/*							*/
void init_iList_vector(iList *v, int length)
{
iList *help;
int i;

help=v;
for (i=1; i<=length; i++)
    {help++; /* first increse - NR-Convention Vectors!! */
     (*help).next  =NULL;
     (*help).prev  =help;
     (*help).number=i;
    }
}

/* .....................................................*/
/*  initialzes iListGrid Tensor component	        */
/* of the range 					*/
/* [imin[1]..imax[1]][imin[2]..imax[2]][imin[3]..imax[3]]*/
/* Initialization: *next =NULL				*/
/*							*/
/* The end and start point of the grid (i.e. imax[] and	*/
/* imin[]) as well as the spacing parameter must be 	*/
/* assigned before calling the function.		*/
/*							*/
void init_iListGrid(iListGrid *grid)
{
int i,j,k;
iList *list;
/* Testings */
if (((*grid).inv_spacing==0.0)||((*grid).spacing==0.0))
   {muerror("Grid spacing is zero in  init_iListGrid()");
   }
if (((*grid).imax==NULL)||((*grid).imax==NULL))
   {muerror("imax or imin not allocated in init_iListGrid()");
   }

/* Initialization */
for (i=(*grid).imin[1]; i<=(*grid).imax[1]; i++)
    for (j=(*grid).imin[2]; j<=(*grid).imax[2]; j++)
	for (k=(*grid).imin[3]; k<=(*grid).imax[3]; k++)
	    {list=&(*grid).root[i][j][k];
	     (*list).next  =NULL;
	     (*list).prev  =NULL;
	     (*list).number=0;
	    }
}

/* .................................................... */
/* Adds a atom to the list				*/
/*							*/
/* INV_SPACING should be assigned before calling this	*/
/* procedure						*/
/*							*/

void add_to_list(iListGrid *grid, iList *list, double *x)
{
int t1, t2, t3;
iList *help, *v;

/* map the atoms on the grid */
t1=(int)rint(x[1]*(*grid).inv_spacing);
t2=(int)rint(x[2]*(*grid).inv_spacing);
t3=(int)rint(x[3]*(*grid).inv_spacing);


if ((*list).number==0) 
   {muwarn("The number is zero in add_to_list().");
   }

/* Testing if the atom is within the grid */
if (  (t1<(*grid).imin[1]) || (t1>(*grid).imax[1])
    ||(t2<(*grid).imin[2]) || (t2>(*grid).imax[2])
    ||(t3<(*grid).imin[3]) || (t3>(*grid).imax[3]))
    {fprintf(stderr,"%8.3f%8.3f%8.3f - %4d%4d%4d\n",
		     x[1], x[2], x[3], t1, t2, t3);
     muerror("Point not in box in add_to_list().");
    }
else{/* redirect the pointer */
     v=&(*grid).root[t1][t2][t3];
     help=(*v).next;
     if (help!=NULL)
        {(*help).prev=list;
	}
     (*v).next=list;
     (*list).next=help;
     (*list).prev=v;
    }
}

/* .................................................... */
/* Removes a entry from the list and redirects 		*/
/* the pointers.					*/
/*							*/
void remove_from_list(iList *list)
{
iList *help;

help=(*list).prev;
(*help).next=(*list).next;

if ((*list).next!=NULL)
   {help=(*list).next;
    (*help).prev=(*list).prev;
   }
/* CHANGE - kann weggelassen werden */
(*list).prev=NULL;
(*list).next=NULL;
}
/* .................................................... */
/* Moves an atom from the old grid cell to a new	*/
/*							*/
/* an alternative but slightly slower version (two 	*/
/* function calls) would be:				*/
/*							*/
/*  remove_from_list(list);				*/
/*  add_to_list(grid, list, x);				*/
/*							*/
/* This should be used if some problem are assumed in	*/
/* nb-list updating					*/
/*							*/

void move_in_list(iListGrid *grid, iList *list, double *x)
{
int t1, t2, t3;
iList *help, *v;

/* first remove the atom from the list */
help=(*list).prev;
(*help).next=(*list).next;
if ((*list).next!=NULL)
   {help=(*list).next;
    (*help).prev=(*list).prev;
   }

/* now reput it in the list  */
/* map the atoms on the grid */
t1=(int)rint(x[1]*(*grid).inv_spacing);
t2=(int)rint(x[2]*(*grid).inv_spacing);
t3=(int)rint(x[3]*(*grid).inv_spacing);


if ((*list).number==0) 
   {muwarn("The number is zero in move_in_list().");
   }

/* Testing if the atom is within the grid */
if (  (t1<(*grid).imin[1]) || (t1>(*grid).imax[1])
    ||(t2<(*grid).imin[2]) || (t2>(*grid).imax[2])
    ||(t3<(*grid).imin[3]) || (t3>(*grid).imax[3]))
    {muwarn("Point not in box in add_to_list().");
     fprintf(stderr,"%8.3f%8.3f%8.3f - %4d%4d%4d\n",
		     x[1], x[2], x[3], t1, t2, t3); 
     (*list).prev=NULL;
     (*list).next=NULL;
    }
else{/* redirect the pointer */
     v=&(*grid).root[t1][t2][t3];
     help=(*v).next;
     if (help!=NULL)
        {(*help).prev=list;
	}
     (*v).next=list;
     (*list).next=help;
     (*list).prev=v;
    }
}
/* .................................................... */
void move_in_list_test(iListGrid *grid, iList *list, double *x)
{
remove_from_list(list);
add_to_list(grid, list, x);
}
/* .................................................... */
/* Fills the list with the given atom coordinates	*/
/*							*/
void fill_iListGrid(iListGrid *grid, iList *list, double **xyz, int atoms)
{
int i;

for (i=1; i<=atoms; i++)
    {add_to_list(grid, &list[i], xyz[i]);
    }
}
/********************************************************/
/* From here on, these are procedures for single linked */
/* neighbor lists					*/
/* .................................................... */
/* Add a atom to a single-linked list			*/
/*							*/	   
void add_to_list2(iListGrid *grid, iList *list, double *x)
{
int t1, t2, t3;
iList *help, *v;

/* map the atoms on the grid */
t1=(int)rint(x[1]*(*grid).inv_spacing);
t2=(int)rint(x[2]*(*grid).inv_spacing);
t3=(int)rint(x[3]*(*grid).inv_spacing);


if ((*list).number==0) 
   {muwarn("The number is zero in add_to_list().");
   }

/* Testing if the atom is within the grid */
if (  (t1<(*grid).imin[1]) || (t1>(*grid).imax[1])
    ||(t2<(*grid).imin[2]) || (t2>(*grid).imax[2])
    ||(t3<(*grid).imin[3]) || (t3>(*grid).imax[3]))
    {fprintf(stderr,"%8.3f%8.3f%8.3f - %4d%4d%4d\n",
		     x[1], x[2], x[3], t1, t2, t3);
     muerror("Point not in box in add_to_list().");
    }
else{/* redirect the pointer */
     v=&(*grid).root[t1][t2][t3];
     help=(*v).next;
     (*v).next=list;
     (*list).next=help;
    }
}

/* .................................................... */
/* fills neighbor list with a single-linked list.	*/
/*							*/
void fill_iListGrid2(iListGrid *grid, iList *list, double **x,  int natoms)
{
int i,j,k;
iList *help;

/* Initialization */
for (i=(*grid).imin[1]; i<=(*grid).imax[1]; i++)
    for (j=(*grid).imin[2]; j<=(*grid).imax[2]; j++)
	for (k=(*grid).imin[3]; k<=(*grid).imax[3]; k++)
	    {help=&(*grid).root[i][j][k];
	     (*help).next  =NULL;
	    }

for (i=1; i<=natoms;i++)
    {add_to_list2(grid, &list[i], x[i]);
    }
}

/* .................................................... */
/* removes an atom from a single-linked list		*/
/*							*/
void remove_from_list2(iListGrid *grid, iList *list, double *x)
{
int t1, t2, t3,i;
iList *prev, *v;

/* map the atoms on the grid */
t1=(int)rint(x[1]*(*grid).inv_spacing);
t2=(int)rint(x[2]*(*grid).inv_spacing);
t3=(int)rint(x[3]*(*grid).inv_spacing);


if ((*list).number==0) 
   {muwarn("The number is zero in add_to_list().");
   }

/* Testing if the atom is within the grid */
if (  (t1<(*grid).imin[1]) || (t1>(*grid).imax[1])
    ||(t2<(*grid).imin[2]) || (t2>(*grid).imax[2])
    ||(t3<(*grid).imin[3]) || (t3>(*grid).imax[3]))
    {fprintf(stderr,"%8.3f%8.3f%8.3f - %4d%4d%4d\n",
		     x[1], x[2], x[3], t1, t2, t3);
     muerror("Point not in box in new_remove_from_list().");
    }
else{/* redirect the pointer */
     v=&(*grid).root[t1][t2][t3];
     if ((*v).next==NULL) 
	{muerror("This should never happen in remove_from_list2().");
	}
     prev=v;
     v=(*v).next;
     i=(*v).number;
     while(i!=(*list).number)
	  {prev=v;
           v=(*v).next;
           i=(*v).number;
	   if (v==NULL) 
	      {muerror("This should never happen in remove_from_list2().");
	      }
	  }
     (*prev).next=(*v).next;
    }
}

