/* Matthias Ullmann 					*/
/* Berlin, 03.10.1995					*/
/* .....................................................*/
#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include <math.h>
#include "util_c.h"
#include "properties.h"
#include "mathematik.h"
#include "constants.h"

/*..............................................................*/
/* all arrays have NR-convention				*/
/* a[i] is a pointer to a[i][0]					*/
/*..............................................................*/

/****************************************************************/
/*  Functions for calculating properties			*/
/****************************************************************/

/*..............................................................*/
/* procedure calculates the geometrical center.			*/

void center(double **xyz, int atomnumber, double *center)
{
int i;
center[1]=0.0;
center[2]=0.0;
center[3]=0.0;

for (i=1; i<=atomnumber; i++)
	{center[1]=center[1]+xyz[i][1];
	 center[2]=center[2]+xyz[i][2];
	 center[3]=center[3]+xyz[i][3];
	}	
center[1]=center[1]/atomnumber;
center[2]=center[2]/atomnumber;
center[3]=center[3]/atomnumber;
}

/*..............................................................*/
/* procedure calculates the center of mass.			*/
/* it returns the total mass.					*/

double center_of_mass(double **xyz, double *mass, int atomnumber, double *masscenter)
{
double total_mass;
int i;
masscenter[1]=0.0;
masscenter[2]=0.0;
masscenter[3]=0.0;

total_mass=0.0;
for (i=1; i<=atomnumber; i++)
	{masscenter[1]=masscenter[1]+(mass[i]*xyz[i][1]);
	 masscenter[2]=masscenter[2]+(mass[i]*xyz[i][2]);
	 masscenter[3]=masscenter[3]+(mass[i]*xyz[i][3]);
	 if (mass[i]>70.0) 
	       {fprintf(stderr,"\n\nWARNING: Large mass - Atom Number %d.\n",i);
	       }
	 if (mass[i]<0.0) 
	       {fprintf(stderr,"\n\nWARNING: Negatve mass - Atom Number %d.\n",i);
	       }
	 if (mass[i]==0.0) 
	       {fprintf(stderr,"\n\nWARNING: Zero mass - Atom Number %d.\n",i);
	       }
	 total_mass=total_mass+mass[i];
	}	
masscenter[1]=masscenter[1]/total_mass;
masscenter[2]=masscenter[2]/total_mass;
masscenter[3]=masscenter[3]/total_mass;
return total_mass;
}

/*..............................................................*/
/* procedure calculates the volume of a protein by summing	*/
/* up the volume of the van der Waals-spheres			*/
/* a warning appears, if the radius is larger than 2.3		*/

double volume(double **xyz, double *radius, int atomnumber)
{
double volume=0.0;
double r3=0.0; /* Radius^3*/
int i;

for (i=1; i<=atomnumber; i++)
	{r3=r3+(radius[i]*radius[i]*radius[i]);
	 if (radius[i]>3.5) 
	       {fprintf(stderr,"\n\nWARNING: Large radius - Atom Number %d.\n",i);
	       }
       	}
volume=((4/3)*(M_PI*r3));
	
return volume;
}

/*..............................................................*/
/* procedure calculates the total charge of a protein 		*/

double charge(double *charge, int atomnumber)
{
double Charge=0.0;
int i;

for (i=1; i<=atomnumber; i++)
	{Charge= Charge + charge[i];
	}	
return Charge;
}

/*..............................................................*/
/* procedure calculates the dipole-vector			*/
/* The protein is virtually shifted to the center of mass.	*/
/* The center of mass must be given to the routine.		*/
/* it returns the dipolmoment in Columb*Angstrom		*/

double dipol(double **xyz, double *charge, int atomnumber, 
	     double *masscenter, double *dipolvec)
{
  int i;
  double dipol;
  
  dipolvec[1]=0.0;
  dipolvec[2]=0.0;
  dipolvec[3]=0.0;
  for (i=1; i<=atomnumber; i++)
          {dipolvec[1]=dipolvec[1]+(charge[i]*(xyz[i][1]-masscenter[1]));
           dipolvec[2]=dipolvec[2]+(charge[i]*(xyz[i][2]-masscenter[2]));
           dipolvec[3]=dipolvec[3]+(charge[i]*(xyz[i][3]-masscenter[3]));
          }
  
  dipol=length(dipolvec,3);
  return dipol;
}	

/*..............................................................*/
/* procedure calculates the gyrationradius 			*/
/* needs the center of mass					*/
double gyrationradius(double **xyz, int atomnumber, double *masscenter)
{
int i;
double gyration=0.0; /* distance from center of mass */

for (i=1; i<=atomnumber; i++)
	{gyration=gyration + sqr_distance(masscenter,xyz[i],3);
	}

return sqrt(gyration/atomnumber);
}	

/*..............................................................*/
/* procedure calcultes the distance of each vector in an array 	*/
/* of vectors xyz from a given point a and return the maximal 	*/
/* distance.							*/
/* for instance: maximal distance of all atoms from the center	*/
/* of mass.							*/
double max_dist(double **xyz, int number, double *a, double *dist)
{
double max_distance = -1.0;
int i;

for (i=1; i<= number; i++)
	{dist[i]=sqrt(sqr_distance(a,xyz[i],3));
	 if (max_distance<dist[i])
		{max_distance=dist[i];
		}
	}
if (max_distance==-1.0)
	{muwarn("Error while calculating distances.\n");
	}
return max_distance;
}

/****************************************************************/
/*  Functions for Coordinate-Manipultions			*/
/****************************************************************/

/*..............................................................*/
/* procedure translates a molecule 				*/

void translation(double **xyz, int atomnumber, double *trans)
{
int i;
for (i=1; i<=atomnumber; i++)
	{xyz[i][1]=trans[1]+xyz[i][1];
	 xyz[i][2]=trans[2]+xyz[i][2];
	 xyz[i][3]=trans[3]+xyz[i][3];
	}	
}

/*..............................................................*/
/* procedure rotates a molecule with respect to the given	*/
/* center of rotation.						*/
/* 	1. Translation of the molecule to the center of		*/
/*	   rotation.						*/
/*	2. Rotation						*/
/* 	3. Back Translation					*/
/* 								*/
/* ATTENTION: The angle vector must be given in rad !!!		*/
/*								*/

void rotation(double **xyz, int atomnumber, 
	      double *center_of_rotation, double *angles)
{
double c[4],s[4];
double rot[4][4];
double old1[4], old2[4];
int i;

/* set up of the rotation matrix */
c[1]=cos(angles[1]);   s[1]=sin(angles[1]);
c[2]=cos(angles[2]);   s[2]=sin(angles[2]);
c[3]=cos(angles[3]);   s[3]=sin(angles[3]);

rot[1][1]=(c[2]*c[3]);
rot[1][2]=(s[1]*s[2]*c[3]) - (c[1]*s[3]);
rot[1][3]=(s[1]*s[3]) + (c[1]*s[2]*c[3]);
rot[2][1]=(c[2]*s[3]);
rot[2][2]=(c[1]*c[3]) + (s[1]*s[2]*s[3]);
rot[2][3]=(c[1]*s[2]*s[3]) - (s[1]*c[3]);
rot[3][1]=-s[2];
rot[3][2]=s[1]*c[2];
rot[3][3]=c[1]*c[2];

for (i=1; i<=atomnumber; i++)
	{/* Translation of the molecule to the center of rotation. */
	 old1[1]=xyz[i][1]-center_of_rotation[1];
	 old1[2]=xyz[i][2]-center_of_rotation[2];
	 old1[3]=xyz[i][3]-center_of_rotation[3];
	 /* Rotation */
	 old2[1]=rot[1][1]*old1[1] + rot[1][2]*old1[2] + rot[1][3]*old1[3];
	 old2[2]=rot[2][1]*old1[1] + rot[2][2]*old1[2] + rot[2][3]*old1[3];
	 old2[3]=rot[3][1]*old1[1] + rot[3][2]*old1[2] + rot[3][3]*old1[3];
	 /* back translation */
	 xyz[i][1]=old2[1]+center_of_rotation[1];
	 xyz[i][2]=old2[2]+center_of_rotation[2];
	 xyz[i][3]=old2[3]+center_of_rotation[3];
	}	

}
/*..............................................................*/
/* procedure rotates a molecule with respect to the given	*/
/* center of rotation.						*/
/* 	1. Translation of the molecule to the center of		*/
/*	   rotation.						*/
/*	2. Rotation						*/
/* 	3. Back Translation					*/
/* The rotation matix is given to the program.			*/

void rotate(double **xyz, int atomnumber, 
	      double *center_of_rotation, double **rot)
{
double old1[4], old2[4];
int i;

for (i=1; i<=atomnumber; i++)
	{/* Translation of the molecule to the center of rotation. */
	 old1[1]=xyz[i][1]-center_of_rotation[1];
	 old1[2]=xyz[i][2]-center_of_rotation[2];
	 old1[3]=xyz[i][3]-center_of_rotation[3];
	 /* Rotation */
	 old2[1]=rot[1][1]*old1[1] + rot[1][2]*old1[2] + rot[1][3]*old1[3];
	 old2[2]=rot[2][1]*old1[1] + rot[2][2]*old1[2] + rot[2][3]*old1[3];
	 old2[3]=rot[3][1]*old1[1] + rot[3][2]*old1[2] + rot[3][3]*old1[3];
	 /* back translation */
	 xyz[i][1]=old2[1]+center_of_rotation[1];
	 xyz[i][2]=old2[2]+center_of_rotation[2];
	 xyz[i][3]=old2[3]+center_of_rotation[3];
	}	

}
/*..............................................................*/
/* procedure rotates a molecule with respect to the given	*/
/* center of rotation.						*/
/* rotates also an additional vector (for instance center of 	*/
/* mass.							*/
/* 	1. Translation of the molecule to the center of		*/
/*	   rotation.						*/
/*	2. Rotation						*/
/* 	3. Back Translation					*/
/* 								*/
/* ATTENTION: The angle vector must be given in rad !!!		*/
/*								*/

void rotation2(double **xyz, int atomnumber, double *vector,
	      double *center_of_rotation, double *angles)
{
double c[4],s[4];
double rot[4][4];
double old1[4], old2[4];
int i;

/* set up of the rotation matrix */
c[1]=cos(angles[1]);   s[1]=sin(angles[1]);
c[2]=cos(angles[2]);   s[2]=sin(angles[2]);
c[3]=cos(angles[3]);   s[3]=sin(angles[3]);

rot[1][1]=(c[2]*c[3]);
rot[1][2]=(s[1]*s[2]*c[3]) - (c[1]*s[3]);
rot[1][3]=(s[1]*s[3]) + (c[1]*s[2]*c[3]);
rot[2][1]=(c[2]*s[3]);
rot[2][2]=(c[1]*c[3]) + (s[1]*s[2]*s[3]);
rot[2][3]=(c[1]*s[2]*s[3]) - (s[1]*c[3]);
rot[3][1]=-s[2];
rot[3][2]=s[1]*c[2];
rot[3][3]=c[1]*c[2];

/* Translation of the molecule to the center of rotation. */
old1[1]=vector[1]-center_of_rotation[1];
old1[2]=vector[2]-center_of_rotation[2];
old1[3]=vector[3]-center_of_rotation[3];
/* Rotation */
old2[1]=rot[1][1]*old1[1] + rot[1][2]*old1[2] + rot[1][3]*old1[3];
old2[2]=rot[2][1]*old1[1] + rot[2][2]*old1[2] + rot[2][3]*old1[3];
old2[3]=rot[3][1]*old1[1] + rot[3][2]*old1[2] + rot[3][3]*old1[3];
/* back translation */
vector[1]=old2[1]+center_of_rotation[1];
vector[2]=old2[2]+center_of_rotation[2];
vector[3]=old2[3]+center_of_rotation[3];
for (i=1; i<=atomnumber; i++)
	{/* Translation of the molecule to the center of rotation. */
	 old1[1]=xyz[i][1]-center_of_rotation[1];
	 old1[2]=xyz[i][2]-center_of_rotation[2];
	 old1[3]=xyz[i][3]-center_of_rotation[3];
	 /* Rotation */
	 old2[1]=rot[1][1]*old1[1] + rot[1][2]*old1[2] + rot[1][3]*old1[3];
	 old2[2]=rot[2][1]*old1[1] + rot[2][2]*old1[2] + rot[2][3]*old1[3];
	 old2[3]=rot[3][1]*old1[1] + rot[3][2]*old1[2] + rot[3][3]*old1[3];
	 /* back translation */
	 xyz[i][1]=old2[1]+center_of_rotation[1];
	 xyz[i][2]=old2[2]+center_of_rotation[2];
	 xyz[i][3]=old2[3]+center_of_rotation[3];
	}	

}

/*..............................................................*/
/* procedure rotates a molecule with respect to the origin.	*/
/*								*/
/*	 Rotation						*/
/* 								*/
/* ATTENTION: The angle vector must be given in rad !!!		*/
/*								*/

void ori_rotation(double **xyz, int atomnumber, double *angles)
{
double c[4],s[4];
double rot[4][4];
double help[4];
int i;

/* set up of the rotation matrix */
c[1]=cos(angles[1]);   s[1]=sin(angles[1]);
c[2]=cos(angles[2]);   s[2]=sin(angles[2]);
c[3]=cos(angles[3]);   s[3]=sin(angles[3]);

rot[1][1]=(c[2]*c[3]);
rot[1][2]=(s[1]*s[2]*c[3]) - (c[1]*s[3]);
rot[1][3]=(s[1]*s[3]) + (c[1]*s[2]*c[3]);
rot[2][1]=(c[2]*s[3]);
rot[2][2]=(c[1]*c[3]) + (s[1]*s[2]*s[3]);
rot[2][3]=(c[1]*s[2]*s[3]) - (s[1]*c[3]);
rot[3][1]=-s[2];
rot[3][2]=s[1]*c[2];
rot[3][3]=c[1]*c[2];

/* Rotation of the vector */
for (i=1; i<=atomnumber; i++)
	{/* Rotation of the array */
	 help[1]=rot[1][1]*xyz[i][1] + rot[1][2]*xyz[i][2] + rot[1][3]*xyz[i][3];
	 help[2]=rot[2][1]*xyz[i][1] + rot[2][2]*xyz[i][2] + rot[2][3]*xyz[i][3];
	 help[3]=rot[3][1]*xyz[i][1] + rot[3][2]*xyz[i][2] + rot[3][3]*xyz[i][3];
	 xyz[i][1]=help[1];
	 xyz[i][2]=help[2];
	 xyz[i][3]=help[3];
	}	

}

/*..............................................................*/
/* procedure rotates a molecule with respect to the origin.	*/
/*								*/
/* rotates also an additional vector (for instance center of 	*/
/* mass.							*/
/*	 Rotation						*/
/* 								*/
/* ATTENTION: The angle vector must be given in rad !!!		*/
/*								*/

void ori_rotation2(double **xyz, int atomnumber, double *vector,
	      double *angles)
{
double c[4],s[4];
double rot[4][4];
double help[4];
int i;

/* set up of the rotation matrix */
c[1]=cos(angles[1]);   s[1]=sin(angles[1]);
c[2]=cos(angles[2]);   s[2]=sin(angles[2]);
c[3]=cos(angles[3]);   s[3]=sin(angles[3]);

rot[1][1]=(c[2]*c[3]);
rot[1][2]=(s[1]*s[2]*c[3]) - (c[1]*s[3]);
rot[1][3]=(s[1]*s[3]) + (c[1]*s[2]*c[3]);
rot[2][1]=(c[2]*s[3]);
rot[2][2]=(c[1]*c[3]) + (s[1]*s[2]*s[3]);
rot[2][3]=(c[1]*s[2]*s[3]) - (s[1]*c[3]);
rot[3][1]=-s[2];
rot[3][2]=s[1]*c[2];
rot[3][3]=c[1]*c[2];

/* Rotation of the vector */
help[1]=rot[1][1]*vector[1] + rot[1][2]*vector[2] + rot[1][3]*vector[3];
help[2]=rot[2][1]*vector[1] + rot[2][2]*vector[2] + rot[2][3]*vector[3];
help[3]=rot[3][1]*vector[1] + rot[3][2]*vector[2] + rot[3][3]*vector[3];
vector[1]=help[1];
vector[2]=help[2];
vector[3]=help[3];

/* Rotation of the array */
for (i=1; i<=atomnumber; i++)
	{help[1]=rot[1][1]*xyz[i][1] + rot[1][2]*xyz[i][2] + rot[1][3]*xyz[i][3];
	 help[2]=rot[2][1]*xyz[i][1] + rot[2][2]*xyz[i][2] + rot[2][3]*xyz[i][3];
	 help[3]=rot[3][1]*xyz[i][1] + rot[3][2]*xyz[i][2] + rot[3][3]*xyz[i][3];
	 xyz[i][1]=help[1];
	 xyz[i][2]=help[2];
	 xyz[i][3]=help[3];
	}	

}

/*..............................................................*/
/* procedure first rotates a molecule, than translates		*/
/* rotation analog to 'rotation'.				*/
/* it also translates the center of rotaion 			*/
/* new coordinates in newxyz					*/

void move(double **oldxyz, int atomnumber, 
	  double *trans, double *center_of_rotation, double *angles, 
	  double **newxyz, double *new_center)
{
double c[4],s[4];
double rot[4][4];
double old1[4], old2[4];
int i;

/* set up of the rotation matrix */
c[1]=cos(angles[1]);   s[1]=sin(angles[1]);
c[2]=cos(angles[2]);   s[2]=sin(angles[2]);
c[3]=cos(angles[3]);   s[3]=sin(angles[3]);

rot[1][1]=(c[2]*c[3]);
rot[1][2]=(s[1]*s[2]*c[3]) - (c[1]*s[3]);
rot[1][3]=(s[1]*s[3]) + (c[1]*s[2]*c[3]);
rot[2][1]=(c[2]*s[3]);
rot[2][2]=(c[1]*c[3]) + (s[1]*s[2]*s[3]);
rot[2][3]=(c[1]*s[2]*s[3]) - (s[1]*c[3]);
rot[3][1]=-s[2];
rot[3][2]=s[1]*c[2];
rot[3][3]=c[1]*c[2];

for (i=1; i<=atomnumber; i++)
	{/* Translation of the molecule to the center of rotation. */
	 old1[1]=oldxyz[i][1]-center_of_rotation[1];
	 old1[2]=oldxyz[i][2]-center_of_rotation[2];
	 old1[3]=oldxyz[i][3]-center_of_rotation[3];
	 /* Rotation */
	 old2[1]=rot[1][1]*old1[1] + rot[1][2]*old1[2] + rot[1][3]*old1[3];
	 old2[2]=rot[2][1]*old1[1] + rot[2][2]*old1[2] + rot[2][3]*old1[3];
	 old2[3]=rot[3][1]*old1[1] + rot[3][2]*old1[2] + rot[3][3]*old1[3];
	 /* back translation and translation*/
	 newxyz[i][1]=trans[1]+old2[1]+center_of_rotation[1];
	 newxyz[i][2]=trans[2]+old2[2]+center_of_rotation[2];
	 newxyz[i][3]=trans[3]+old2[3]+center_of_rotation[3];
	}	
/* translation of the center of rotation */
new_center[1]=center_of_rotation[1] + trans[1];
new_center[2]=center_of_rotation[2] + trans[2];
new_center[3]=center_of_rotation[3] + trans[3];
}

/*..............................................................*/
/* Rotation around a arbitrary axis				*/
/* The rotation axis *must* be an unit vector			*/
/* Idea D. Hoffmann						*/
/*								*/

void rotate_around_vec(double **xyz, int atomnumber, double *axis, double angle)
{
int i;
double cos_ang, sin_ang;
double scalar_prod;
double vector_prod[4];

cos_ang=cos(angle);
sin_ang=sin(angle);

for (i=1; i<=atomnumber; i++)
	{scalar_prod=ScalarProduct(xyz[i],axis);
	 VectorProduct(xyz[i],axis,vector_prod);
	 xyz[i][1]=scalar_prod*axis[1] + (xyz[i][1]-scalar_prod*axis[1])*cos_ang - vector_prod[1]*sin_ang;
	 xyz[i][2]=scalar_prod*axis[2] + (xyz[i][2]-scalar_prod*axis[2])*cos_ang - vector_prod[2]*sin_ang;
	 xyz[i][3]=scalar_prod*axis[3] + (xyz[i][3]-scalar_prod*axis[3])*cos_ang - vector_prod[3]*sin_ang;
	}
}

/*..............................................................*/
/* The procedure calclutes the out-of-plane-distance of the 	*/
/* point d, from a plane which is spaned from three atoms a,b	*/
/* and c. The result is returned to the calling program		*/
/*								*/

double out_of_plane(double *a, double *b, double *c, double *d)
{
double h, cp[4]; /* height, cross product */
double aa[4], bb[4], cc[4];

SubVector(d,a,aa);
SubVector(b,a,bb);
SubVector(c,a,cc);

VectorProduct(cc,bb,cp);
h=LENGTH(cp);

h=ScalarProduct(aa,cp)/h;

return fabs(h);
}

/*.................................................*/
/* This rountine performes the BIOMT transfomation */
/* to create the biologically active ternary       */
/* structure and the SMTRY transformation to create*/
/* the crystal environment                         */
/*                                                 */
void biomt(double **xyz, int n, double **matrix)
{
int i,k,l;
double y[4];

for (i=1; i<=n; i++)
    {for (k=1; k<=3; k++)
         {y[k]=matrix[k][4];
          for (l=1; l<=3; l++)
	      {y[k]=y[k] + matrix[k][l]*xyz[i][l];
	      }
	 }
     for (l=1; l<=3; l++)
         {xyz[i][l]=y[l];
         }
    }
}

/****************************************************************/
/* some stuff							*/
/****************************************************************/
/* .................................................................... */
/* returns the atom id number in aminoacid aa_no in mus with the atom	*/
/* id given as a char paameter to the function 				*/
int get_no(MUS_atom *mus, int aa_no, char *atom_id)
{
int i,j,s;
int back=0; 

s=(*mus).aa[aa_no][1];
for (i=0; i<(*mus).aa[aa_no][2]; i++)
    {if (0==strcmp(atom_id,(*mus).atomid[s+i]))
	{if (back==0)
	    {back= (*mus).atomno[s+i];
	    }
	 else {muwarn("Atom ID appears twice in get_no().");
    	       for (j=0;j<(*mus).aa[aa_no][2];j++)
        	   {fprint_PQRatom(STDERR,mus,s+j);
		   }
	       fprintf(STDERR,"Aminoacid %d AtomID '%s'\n",aa_no, atom_id);
	      }
	}
    }
if (back==0)
   {muwarn("Atom ID does not appear in get_no().");
    for (j=0;j<(*mus).aa[aa_no][2];j++)
        {fprint_PQRatom(STDERR,mus,s+j);
	}
    fprintf(STDERR,"Aminoacid %d AtomID '%s'\n",aa_no, atom_id);
   }

return back;
}

/* .................................................................... */
/* returns the atom id number in aminoacid aa_no in mus with the atom	*/
/* id given as a char paameter to the function 				*/
/*									*/
/* it searches for a atom wich matches the following three criteria	*/
/* 		1. Atom ID						*/
/*		2. Residue ID						*/
/*		3. Residue Number					*/
/*									*/
int get_no2(MUS_atom *mus, int aa_no, char *res_id, char *atom_id)
{
int i;
int back=0; 
int count=0;

for (i=1; i<=(*mus).number_of_atoms; i++)
    {if (0==strcmp(atom_id,(*mus).atomid[i]))
	{if (0==strcmp(res_id,(*mus).residue[i]))
	    {if (aa_no==(*mus).res_no[i])
	        {back=i;
		 count++;
	        }
	    }
	}
    }
if ((back==0)||(count!=1))
   {muwarn("Atom ID does not (or too often) appear in get_no().");
    fprintf(STDERR,"Aminoacid-No %d Aminoacid '%s' AtomID '%s'\n",
		   aa_no, res_id, atom_id);
   }

return back;
}
