/* Matthias Ullmann, Berlin 10.10.1994					*/
/* Enthaelt einige Prozeduren zum generieren von Zufallszahlen.		*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "random.h"
#include "constants.h"


/*..............................................................................*/
/* It returns a random vector with components taken from a Gauss distribution	*/
/* with zero mean and a given deviation.					*/
void random_vector(double *vector, double deviation, long *seed)
{
vector[1]=deviation*gasdev(seed);
vector[2]=deviation*gasdev(seed);
vector[3]=deviation*gasdev(seed);
}


/*..............................................................................*/
/* Function returns normally distributed deviate with zero mean and unit 	*/
/* variance, using Ran as source of uniform deviates.				*/
/* Taken Numerical Recipes. slidely modified. pg.289				*/
/* 	1. float --> double							*/
/*	2. ran1 --> ran2							*/
/* The number idum should be a negative integer, to make the first if true.	*/

double gasdev(long *idum)
{
        double Ran(long *idum);
        static int iset=0;
        static double gset;
        double fac,rsq,v1,v2;

        if  (iset == 0) {
                do {
                        v1=2.0*Ran(idum)-1.0;
                        v2=2.0*Ran(idum)-1.0;
                        rsq=v1*v1+v2*v2;
                } while (rsq >= 1.0 || rsq == 0.0);
                fac=sqrt(-2.0*log(rsq)/rsq);
                gset=v1*fac;
                iset=1;
                return v2*fac;
        } else {
                iset=0;
                return gset;
        }
}
/****************************************************************/
/* Metropolis algorithm						*/
/* ............................................................ */
/* metropolis returns '1' if the step is accepted		*/
/* delta_energy= energy_new - energy_old			*/
/*								*/
/* In order to avoid an underflow error in exp(), a constant was*/
/* defined in energy.h. No error message is issued.		*/

int metropolis(double delta_energy, double temp, long *seed)
{
double help;
/* fprintf(stdout,"Delta Energy: %lf\n",delta_energy);*/
/*fprintf(stdout,"Delta Energy: %e\n",-delta_energy/(temp*R_GAS));*/


if (delta_energy < 0.0) 
   {return ACCEPTED; /* 1 */
   }
else {help=-delta_energy/(temp*R_GAS);
      if (help<TOO_SMALL)
	 {return REJECTED;
	 }
      else {if ((temp!=0.0) && ((Ran(seed)) < exp(help)))
	       {return ACCEPTED; /* 1 */;
	       }
	    else {return REJECTED;
		 }
	   }
     }
}
/* ............................................................ */
/* metropolis returns '1' if the step is accepted		*/
/* delta_energy= energy_new - energy_old			*/
/*								*/
/* In order to avoid an underflow error in exp(), a constant was*/
/* defined in energy.h. No error message is issued.		*/
/*								*/
/* uses beta instead of the temparature.			*/
/*								*/
int metropolis2(double delta_energy, double beta, long *seed)
{
double help;


if (delta_energy < 0.0) 
   {return ACCEPTED; /* 1 */
   }
else {help=-delta_energy/beta;
      if (help<TOO_SMALL)
	 {return REJECTED;
	 }
      else {if ((beta!=0.0) && ((Ran(seed)) < exp(help)))
	       {return ACCEPTED; /* 1 */;
	       }
	    else {return REJECTED;
		 }
	   }
     }
}

/* aus NR
int metropolis(double delta_energy, double temp, long *seed)
{
return delta_energy < 0.0 || Ran(seed) < exp(-delta_energy/(temp*R_GAS));
}
*/
/********************************************************************************/
/* Random number generation - Numerical Recipies				*/
/********************************************************************************/

/* see Numerical Recipies pg. 274-328						*/

/*..............................................................................*/
/* Minimal random number generator of Park and Miller.				*/
/*										*/
/*
#define IA 16807
#define IM 2147483647
#define AM (1.0/IM)
#define IQ 127773
#define IR 2836
#define MASK 123459876

float ran0(long *idum)
{
        long k;
        float ans;

        *idum ^= MASK;
        k=(*idum)/IQ;
        *idum=IA*(*idum-k*IQ)-IR*k;
        if (*idum < 0) *idum += IM;
        ans=AM*(*idum);
        *idum ^= MASK;
        return ans;
}
#undef IA
#undef IM
#undef AM
#undef IQ
#undef IR
#undef MASK
*/
/*..............................................................................*/
/* Minimal random number generator of Park and Miller with Bays-Durham shuffle. */
/*										*/
/*
#define IA 16807
#define IM 2147483647
#define AM (1.0/IM)
#define IQ 127773
#define IR 2836
#define NTAB 32
#define NDIV (1+(IM-1)/NTAB)
#define EPS 1.2e-7
#define RNMX (1.0-EPS)

float ran1(long *idum)
{
        int j;
        long k;
        static long iy=0;
        static long iv[NTAB];
        float temp;

        if (*idum <= 0 || !iy) {
                if (-(*idum) < 1) *idum=1;
                else *idum = -(*idum);
                for (j=NTAB+7;j>=0;j--) {
                        k=(*idum)/IQ;
                        *idum=IA*(*idum-k*IQ)-IR*k;
                        if (*idum < 0) *idum += IM;
                        if (j < NTAB) iv[j] = *idum;
                }
                iy=iv[0];
        }
        k=(*idum)/IQ;
        *idum=IA*(*idum-k*IQ)-IR*k;
        if (*idum < 0) *idum += IM;
        j=iy/NDIV;
        iy=iv[j];
        iv[j] = *idum;
        if ((temp=AM*iy) > RNMX) return RNMX;
        else return temp;
}
#undef IA
#undef IM
#undef AM
#undef IQ
#undef IR
#undef NTAB
#undef NDIV
#undef EPS
#undef RNMX
*/
/*..............................................................................*/
/* Long period (>2*10^18) random number genertator of L'Ecuyer with BAYES-	*/
/* DURHAM shuffle and added safeguards. Returns a uniform deviate between 0.0	*/
/* and 1.0 (exclusive the endpoint values). Call with idum a negative integer	*/
/* to initialize; thereafter, do not alter idum between  successive deviates 	*/
/* in a sequence. RNMX should approximate the largest floating value that is 	*/
/* less than 1.0.								*/
/* Taken from Numerical Recipes. slidely modified. pg.282			*/
/* 	1. float --> double							*/
/* The number idum should be a negative integer, to make the first if true.	*/
/*										*/

#define IM1 2147483563
#define IM2 2147483399
#define AM (1.0/IM1)
#define IMM1 (IM1-1)
#define IA1 40014
#define IA2 40692
#define IQ1 53668
#define IQ2 52774
#define IR1 12211
#define IR2 3791
#define NTAB 32
#define NDIV (1+IMM1/NTAB)
#define EPS 1.2e-7
#define RNMX (1.0-EPS)

double ran2(long *idum)
{
        long j;
        long k;
        static long idum2=123456789;
        static long iy=0;
        static long iv[NTAB];
        double temp;

        if (*idum <= 0) {
                if (-(*idum) < 1) *idum=1;
                else *idum = -(*idum);
                idum2=(*idum);
                for (j=NTAB+7;j>=0;j--) {
                        k=(*idum)/IQ1;
                        *idum=IA1*(*idum-k*IQ1)-k*IR1;
                        if (*idum < 0) *idum += IM1;
                        if (j < NTAB) iv[j] = *idum;
                }
                iy=iv[0];
        }
        k=(*idum)/IQ1;
        *idum=IA1*(*idum-k*IQ1)-k*IR1;
        if (*idum < 0) *idum += IM1;
        k=idum2/IQ2;
        idum2=IA2*(idum2-k*IQ2)-k*IR2;
        if (idum2 < 0) idum2 += IM2;
        j=iy/NDIV;
        iy=iv[j]-idum2;
        iv[j] = *idum;
        if (iy < 1) iy += IMM1;
        if ((temp=AM*(double)iy) > RNMX) return RNMX;
        else return temp;
}
#undef IM1
#undef IM2
#undef AM
#undef IMM1
#undef IA1
#undef IA2
#undef IQ1
#undef IQ2
#undef IR1
#undef IR2
#undef NTAB
#undef NDIV
#undef EPS
#undef RNMX


/*..............................................................................*/
/*
#define MBIG 1000000000
#define MSEED 161803398
#define MZ 0
#define FAC (1.0/MBIG)

float ran3(long *idum)
{
        static int inext,inextp;
        static long ma[56];
        static int iff=0;
        long mj,mk;
        int i,ii,k;

        if (*idum < 0 || iff == 0) {
                iff=1;
                mj=MSEED-(*idum < 0 ? -*idum : *idum);
                mj %= MBIG;
                ma[55]=mj;
                mk=1;
                for (i=1;i<=54;i++) {
                        ii=(21*i) % 55;
                        ma[ii]=mk;
                        mk=mj-mk;
                        if (mk < MZ) mk += MBIG;
                        mj=ma[ii];
                }
                for (k=1;k<=4;k++)
                        for (i=1;i<=55;i++) {
                                ma[i] -= ma[1+(i+30) % 55];
                                if (ma[i] < MZ) ma[i] += MBIG;
                        }
                inext=0;
                inextp=31;
                *idum=1;
        }
        if (++inext == 56) inext=1;
        if (++inextp == 56) inextp=1;
        mj=ma[inext]-ma[inextp];
        if (mj < MZ) mj += MBIG;
        ma[inext]=mj;
        return mj*FAC;
}
#undef MBIG
#undef MSEED
#undef MZ
#undef FAC
*/
/*..............................................................................*/
/*
float ran4(long *idum)
{
        void psdes(unsigned long *lword, unsigned long *irword);
        unsigned long irword,itemp,lword;
        static long idums = 0;
#if defined(vax) || defined(_vax_) || defined(__vax__) || defined(VAX)
        static unsigned long jflone = 0x00004080;
        static unsigned long jflmsk = 0xffff007f;
#else
        static unsigned long jflone = 0x3f800000;
        static unsigned long jflmsk = 0x007fffff;
#endif

        if (*idum < 0) {
                idums = -(*idum);
                *idum=1;
        }
        irword=(*idum);
        lword=idums;
        psdes(&lword,&irword);
        itemp=jflone | (jflmsk & irword);
        ++(*idum);
        return (*(float *)&itemp)-1.0;
}
*/
/*..............................................................................*/
/* 
Daniel Hoffmann
Diese Routine liefert Zufallszahlen auf dem Intervall [0;Pi] deren 
Haeufigkeit proportional zu sin(x) ist. Ersetze drand48 durch den
Zufallszahlengenerator Deiner Wahl.
*/

double RandomSinus(long *idum)
{
  double new2, rho;
  int NotAccepted;

  NotAccepted=1;
  while (NotAccepted) {
    rho=sin(M_PI*(new2=Ran(idum)));
    if (Ran(idum)>1.0-rho)
      NotAccepted=0;
  }

  return (M_PI*new2);
}
