//
// C++ Implementation: mit_surface
//
// Description:
//
//
// Author: Mirco <mirco@Mirco-Laptop>, (C) 2008
//
// Copyright: See COPYING file that comes with this distribution
//
//
#include "mit_surface.h"
/** Constructor

Calculating the dot surface and everything necessary for this calculation
\param atomsurf
\param xyz Coordinates of the protein atoms
\param radius Atom radii
\param natom Number of atoms
\param densit
\param prob Probe sphere radius
\param surfflag Recalculate the dot surface
\todo Should be commented by Matthias!
*/
mit_surface::mit_surface(double *atomsurf, double **xyz, double *radius, int natoms, int densit, double prob, int surfflag) {
  int i, j;
  double max[4], min[4], off_set, srf, *rad;
  double len;
  int itest;
  Surf surf;
  iListGrid atom_list;
  iList *atoms;
  char **flag;  /* [1..natom][1..n_dot] INSIDE - OUTSIDE */


  rad = dvector(1, natoms);
  for (i = 1; i <= natoms; i++)
    rad[i] = radius[i] + prob;

  /* determinte which tessellation type should be used */
  itest = tessellation_type(densit);
  /* calculate the tesselation of a unit sphere */
  switch (itest) {
    case UNSP_ICO_ARC:  fprintf(stdout, "Use tessellation type II\n"),
      ico_dot_arc(densit, &surf);
      fprintf(stdout, "I use %d dots\n", surf.n_dot);
      break;
    case UNSP_ICO_DOD:  fprintf(stdout, "Use tessellation type I\n");
      ico_dot_dod(densit, &surf);
      fprintf(stdout, "I use %d dots\n", surf.n_dot);
      break;
    default:            cerr << "Don't know this tessellation type" << endl;
  }

  /* ============================================================= */
  /* renormalize in order to be sure that it is a unit sphere       */
  for (j = 1; j <= surf.n_dot; j++) {
    len = sqrt(SqrLength(surf.xus[j]));
    surf.xus[j][1] = surf.xus[j][1] / len;
    surf.xus[j][2] = surf.xus[j][2] / len;
    surf.xus[j][3] = surf.xus[j][3] / len;
  }
  /* ============================================================= */
  /* sort the dots in a Grid linked cubes=cubes in one dimension   */
  surf.cubes = 5;
  unitsphere_to_grid(&surf);


  /* ============================================================ */
  /* allocate and initilize the flags with OUTSIDE */
  /* look for the needed grid dimension */
  atom_list.imax = ivector(1, 3);
  atom_list.imin = ivector(1, 3);

  off_set = 10.0;
  atom_list.spacing = 2.0 * max_element(rad, natoms);
  atom_list.inv_spacing = 1.0 / atom_list.spacing;
  min_max_coord(min, max, xyz, natoms);
  atom_list.inv_spacing = 1.0 / atom_list.spacing;
  atom_list.imax[1] = (int) ceil((max[1] + off_set) * atom_list.inv_spacing);
  atom_list.imax[2] = (int) ceil((max[2] + off_set) * atom_list.inv_spacing);
  atom_list.imax[3] = (int) ceil((max[3] + off_set) * atom_list.inv_spacing);
  atom_list.imin[1] = (int) floor((min[1] - off_set) * atom_list.inv_spacing);
  atom_list.imin[2] = (int) floor((min[2] - off_set) * atom_list.inv_spacing);
  atom_list.imin[3] = (int) floor((min[3] - off_set) * atom_list.inv_spacing);
  atom_list.look = 1;
  /* make the grid spacing to odd integer, if they are not odd integer */
  odding_down(atom_list.imin, 3);
  odding_up(atom_list.imax, 3);

  /* allocation */
  flag = charmatrix(1, natoms, 1, surf.n_dot);
  atoms = iList_vector(1, natoms);
  atom_list.root = iList_3tensor(atom_list.imin[1], atom_list.imax[1],
                                 atom_list.imin[2], atom_list.imax[2],
                                 atom_list.imin[3], atom_list.imax[3]);

  for (j = 1; j <= natoms; j++) {
    init_charvector(flag[j], surf.n_dot, OUTSIDE);
    atomsurf[j] = 0.0;
  }
  init_iList_vector(atoms, natoms);
  init_iListGrid(&atom_list);
  fill_iListGrid(&atom_list, atoms, xyz, natoms);
  srf = calculate_surface(atomsurf, flag, xyz, rad, natoms,
                          &surf, &atom_list, atoms);

  if (surfflag == 1) {/* calculate the dot surface */
    /* count the surface dots and allocate memory for the surface dots */
    dotsurf.ndots = 0;
    for (i = 1; i <= natoms; i++)
      for (j = 1; j <= surf.n_dot;j++)
        if (flag[i][j] == OUTSIDE)dotsurf.ndots++;

    dotsurf.dots = dmatrix(1, dotsurf.ndots, 1, 3);
    /* do the calculation */
    calculate_dotsurface(&dotsurf, flag, &surf, xyz, rad, natoms);
    save_atom_radii_with_points(&dotsurf, flag, &surf, rad, natoms);
  }

  /* release memory */
  free_i4tensor(surf.dot_no, surf.imin[1], surf.imax[1],
                surf.imin[2], surf.imax[2],
                surf.imin[3], surf.imax[3],
                0, surf.nmax);
  free_dmatrix(surf.xus, 1, surf.n_dot, 1, 3);
  free_dmatrix(surf.help, 1, surf.n_dot, 1, 3);
  free_ivector(surf.imin, 1, 3);
  free_ivector(surf.imax, 1, 3);
  free_dvector(rad, 1, natoms);
  free_charmatrix(flag, 1, natoms, 1, surf.n_dot);
  free_iList_vector(atoms, 1, natoms);
  free_iList_3tensor(atom_list.root, atom_list.imin[1], atom_list.imax[1],
                     atom_list.imin[2], atom_list.imax[2],
                     atom_list.imin[3], atom_list.imax[3]);
  free_ivector(atom_list.imax, 1, 3);
  free_ivector(atom_list.imin, 1, 3);
  surface_in_A = srf;
}


void mit_surface::print_dotsurf_pqr(ofstream &out) {
  for (int i = 1; i <= dotsurf.ndots;i++) {
    write_line_pdb(out, "ATOM", i, "DUM", "DUM", 1, dotsurf.dots[i][1], dotsurf.dots[i][2], dotsurf.dots[i][3], 1.0, 1.0);
  };
}

/** Save the surface per dot

For a calculation of the dot surface from its points (especially after isolating surface point clusters) we need the surface assiciated with each dot.
*/
void mit_surface::save_atom_radii_with_points(DotSurf *dotsurf, char **flag, Surf *surf, double *radius, int natoms) {
  int k = 0;
  surface_per_dot.init_arr(1, (*dotsurf).ndots, 0);
  for (int i = 1; i <= natoms; i++) {
    for (int j = 1; j <= (*surf).n_dot;j++) {
      if (flag[i][j] == OUTSIDE) {
        k++;
        if (k > (*dotsurf).ndots)
          {muerror("Too many dots in calculate_dotsurface()");}
        surface_per_dot(k) = (4.0 * util_pi * radius[i] * radius[i]) / (*surf).n_dot;
      } else {
        if (flag[i][j] != INSIDE) {
          muwarn("Unknown flag in calculate_dotsurface().");
          fprintf(stderr, "%d  %d  %c %d\n", i, j, flag[i][j], (int)flag[i][j]);;
        }
      }
    }
  }
}


/** Cluster a dotsurface

Cluster a dot surface. Clusters are isolated if no point of the cluster is closer than dist to any point outside the cluster. For details about the clustering algorithm, see search_cav().
\param dist maximum Distance between points in one cluster
\param dotgrid The dotsurface
\return ngraph Number of clusters found

*/
int mit_surface::cluster(double dist, iListGrid &dotgrid) {
  int x, y, z;
  iList *help;
  int ii;
  adv_doubly_LL<spoint> all_points;
  for (ii = 1; ii <= dotsurf.ndots; ii++) {
    spoint *n = new spoint;
    n->n = ii;
    all_points.add_tail(n);
  }

  ngraph = 0;
  double_LL<spoint> li;
  spoint *cur;
  surface_clusters.init_arr(0, dotsurf.ndots, 0);
  for (int i = 1; i <= dotsurf.ndots; i++) {
    if (surface_clusters(i) == 0) {
      li.add();
      li.tail->n = i;
      ngraph++;
      //cout << "Graph " << ngraph << " started" << endl;
      int nelements = 0;
      while (li.head != NULL) {
        cur = li.get_last();
        surface_clusters(cur->n) = ngraph;
        x = (int)rint(dotsurf.dots[cur->n][1] * dotgrid.inv_spacing);
        y = (int)rint(dotsurf.dots[cur->n][2] * dotgrid.inv_spacing);
        z = (int)rint(dotsurf.dots[cur->n][3] * dotgrid.inv_spacing);
        // Check in sourrounding boxes
        for (int x2 = x - 2; x2 <= x + 2; x2++) {
          for (int y2 = y - 2; y2 <= y + 2; y2++) {
            for (int z2 = z - 2; z2 <= z + 2; z2++) {
              if (x2 >= dotgrid.imin[1] && x2 <= dotgrid.imax[1] &&
                  y2 >= dotgrid.imin[2] && y2 <= dotgrid.imax[2] &&
                  z2 >= dotgrid.imin[3] && z2 <= dotgrid.imax[3]) {
                help = &dotgrid.root[x2][y2][z2];
                while ((*help).next != NULL) {
                  help = (*help).next;
                  if (surface_clusters(help->number) == 0 && get_dist(dotsurf.dots[cur->n][1], dotsurf.dots[cur->n][2], dotsurf.dots[cur->n][3], dotsurf.dots[(*help).number][1], dotsurf.dots[(*help).number][2], dotsurf.dots[(*help).number][3]) <= dist) {
                    surface_clusters((*help).number) = ngraph;
                    li.add();
                    li.tail->n = (*help).number;
                  }
                }
              }
            };
          };
        };
        nelements++;
        delete cur;
      };
    };
  }
  return ngraph;
}

/** Output surface without clusters

*/
void mit_surface::print_clusters_pqr_surf() {
  char fname[100];
  sprintf(fname, "surface_clusters_surf.pdb");
  ofstream out(fname);
  int nprinted = 0;
  int nres = 1;
  for (int i = 1; i <= dotsurf.ndots;i++) {
    nprinted++;
    if (surface_clusters(i) == 1) {
      if (nprinted > 500) {
        nres++;
        nprinted = 0;
      }
      write_line_pdb(out, "ATOM", i, "DUM", "DUM", nres, dotsurf.dots[i][1], dotsurf.dots[i][2], dotsurf.dots[i][3], 1.0, 1.0);
    }
  };
}
/** Output surface clusters without protein surface(

*/
void mit_surface::print_clusters_pqr_cav() {
  char fname[100];
  sprintf(fname, "surface_clusters_cav.pdb");
  ofstream out(fname);
  int natom = 1;
  for (int ii = 2; ii <= ngraph; ii++) {
    for (int i = 1; i <= dotsurf.ndots;i++) {
      if (surface_clusters(i) == ii) {
        write_line_pdb(out, "ATOM", natom, "DUM", "DUM", surface_clusters(i), dotsurf.dots[i][1], dotsurf.dots[i][2], dotsurf.dots[i][3], 1.0, 1.0);
        natom++;
      };
    };
  }
}

/** Print surface in A^2 for all clusters

*/
void mit_surface::print_clusters_surfaces() {
  vec<double> tmp(1, ngraph, 0);
  for (int i = 1; i <= dotsurf.ndots;i++) {
    tmp(surface_clusters(i)) += surface_per_dot(i);
  };
  for (int i = 1; i <= ngraph; i++) {
    cout << "Surface of cluster " << i  << " " << tmp(i) << endl;
  };
}
