//
// C++ Implementation: monte_carlo_run
//
// Description:
//
//
// Author: Mirco <mirco@Mirco-Laptop>, (C) 2008
//
// Copyright: See COPYING file that comes with this distribution
//
//
#include "monte_carlo_run.h"
/** The whole MC magic

Running the whole MC calculations. Defines a box, calculates the volume of this box. Chooses random points within this box, evaluates, if these points are inside the atom volume, inside voids or in the solvent. For the distribution of points the volume of the protein as well of the solvent and the void ist calculated.
\param atomgrid The coordinates of the protein atoms mapped on a grid used as a neighbourlist
\param dotgrid The previously calculated dotsurface mapped on a grid, used as neighbour list
\param solvgrid Storing the MC points mappen on a grid.
\param p The protein atoms
\param dotsurf The dot surface
\param seed The MC seed
\return double protein volume
*/
double monte_carlo_run::volume_mc(iListGrid *atomgrid, iListGrid *dotgrid, iGrid  *solvgrid, MUS_atom *p, DotSurf *dotsurf, long *seed) {

  double max[4], min[4], length[4], dumm[4];
  int i, loc, loc2, iprt = 1, nmc_A;
  int n_atom = 0, n_void = 0, n_prob = 0, n_solv = 0;
  double boxvol, v1 = 0, v2 = 0, v3 = 0, v4 = 0, dv = DBL_MAX;
  double prob2;
  int ipt[4];

  prob2  = vars.prob * vars.prob;

  max[1] = vars.max[1] + vars.maxrad;
  max[2] = vars.max[2] + vars.maxrad;
  max[3] = vars.max[3] + vars.maxrad;
  min[1] = vars.min[1] - vars.maxrad;
  min[2] = vars.min[2] - vars.maxrad;
  min[3] = vars.min[3] - vars.maxrad;
  length[1] = max[1] - min[1];
  length[2] = max[2] - min[2];
  length[3] = max[3] - min[3];
  boxvol = length[1] * length[2] * length[3];


  nmc_A = (int) ceil((vars.nmc * boxvol) / (*solvgrid).spacing);

  if (vars.blab > 2) {
    /* some more output */
    fprintf(stdout, "The box has the following dimensions:\n");
    fprintf(stdout, "min: %8.3f %8.3f %8.3f \n", min[1], min[2], min[3]);
    fprintf(stdout, "max: %8.3f %8.3f %8.3f \n", max[1], max[2], max[3]);
    fprintf(stdout, "len: %8.3f %8.3f %8.3f \n", length[1], length[2], length[3]);
    cout << "solvegrid min: " << (*solvgrid).imin[1] << " " << (*solvgrid).imin[2] << " " << (*solvgrid).imin[3] << " " << endl;
    cout << "solvegrid max: " << (*solvgrid).imax[1] << " " << (*solvgrid).imax[2] << " " << (*solvgrid).imax[3] << " " << endl;
    fprintf(stdout, "box volume: %8.3f \n", boxvol);
  }
  /* square the radii and the radii+ prob */
  vec<double> r2(1, (*p).number_of_atoms);
  vec<double> pr2(1, (*p).number_of_atoms);
  for (i = 1; i <= (*p).number_of_atoms; i++) {
    r2(i)  = (*p).radius[i] * (*p).radius[i];
    pr2(i) = ((*p).radius[i] + vars.prob) * ((*p).radius[i] + vars.prob);
  }


  (*atomgrid).look = (int)ceil((vars.maxrad + vars.prob) * (*atomgrid).inv_spacing);
  (*dotgrid).look  = (int)ceil(vars.prob * (*dotgrid).inv_spacing);
  if (vars.blab > 1) {
    fprintf(stdout, "Volume  %10s %16s %16s %16s\n", "step", "atom", "atom+void", "dv");
  }
  /* now do the Monte Carlo */
  for (i = 1; i <= nmc_A; i++) {
    dumm[1] = length[1] * ran2(seed) + min[1];
    dumm[2] = length[2] * ran2(seed) + min[2];
    dumm[3] = length[3] * ran2(seed) + min[3];
    // Map the point to the grid
    if (0 != pt_to_cell2(ipt, dumm, solvgrid)) {
      cout << ipt[1] << " " << ipt[2] << " " << ipt[3] << endl;
      cerr << "pt outside the solv grid in main()" << endl;
      exit(0);
    }
    // Check if the point is inside the atom volume
    loc = pt_inside_atom(dumm, atomgrid, (*p).xyz, pr2, r2);
    /* loc = pt_inside_atom_slow(dumm, p.number_of_atoms, p.xyz, pr2, r2); */
    if (1 == loc) {
      n_atom++;
    } else if (-1 == loc) {
      /* check if pt is in small voids */
      n_prob++;
      loc2 = pt_inside_void(dumm, dotgrid, (*dotsurf).dots, prob2);
      if (loc2 == 1) {
        n_void++;
      } else {
        n_solv++;
        (*solvgrid).grid[ipt[1]][ipt[2]][ipt[3]]++;
      }
    } else if (0 == loc) {
      n_solv++;
      (*solvgrid).grid[ipt[1]][ipt[2]][ipt[3]]++;
    } else if (-100 == loc) {
      cerr << "Problem in volume_mc()" << endl;
      exit(0);
    } else {
      cerr << "Serious Problem in volume_mc()" << endl;
      exit(0);
    }

    /* write the output */
    v2 = boxvol * (double)(n_atom + n_void) / i; // Molecular Volume
    v3 = boxvol * (double)(n_atom + n_void + n_prob) / i; // Solvent accessible Volume
    v4 = boxvol * (double)(n_atom) / i; // VdW Volume
    if (vars.blab > 1) {
      if (iprt == vars.nprt) {
        v2 = boxvol * (double)(n_atom + n_void) / i;
        dv = fabs(v1 - v2);
        v1 = v2;
        fprintf(stdout, "VOLUME  %10d %16.6f %16.6f %16.6f\n",
                i, boxvol*(double)n_atom / i, v2, dv);
        iprt = 0;
      }
    }
    iprt++;
  }
  if (vars.blab >= 1) {
    cout << "Monte Carlo data: " << endl;
    cout << "Monte Carlo Steps: " << nmc_A << endl;
    cout << "Inside atoms: " << n_atom << endl;
    cout << "Inside voids: " << n_void << endl;
    cout << "Inside probe sphere and not inside void or atom: " << n_prob << endl;
    cout << "Inside solvent: " << n_solv << endl;
    cout << "Protein Volume: " << v2 << endl;
    cout << "Solvent Accessible Volume: " << v3 << endl;
    cout << "VdW Volume: " << v4 << endl;
  };
  if (vars.blab > 2) {
    fprintf(stdout, "n_atom  = %10d\n", n_atom);
    fprintf(stdout, "n_void  = %10d\n", n_void);
    fprintf(stdout, "n_prob  = %10d\n", n_prob);
    fprintf(stdout, "n_solv  = %10d\n", n_solv);
    v2 = boxvol * (double)(n_atom + n_void) / i;
    fprintf(stdout, "Final Volume  %16.6f %16.6f\n",
            boxvol*(double)n_atom / i, v2);
  }
  return v2;
}


/** Check if a point is inside the atom volume

Checks if a MC point is inside the protein atoms volume. Not all atoms are checked but a neighbour list is evaluated to check only for nearby atoms. This increases the speed of this function a lot
\param pt The point to evaluate
\param atomsgrid Protein atoms mapped on a grid used as neighbourlist
\param xyz Coordinates of the atoms
\param pr2 (atom radius + probe sphere radius)^2
\param r2 atom radius^2
\return int 1 means point is inside the atom volume, -1 point is between atom an probesphere, 0 means point is in solvent
*/
int monte_carlo_run::pt_inside_atom(double *pt, iListGrid *atomgrid, double **xyz, vec<double> &pr2, vec<double> &r2) {
  int iptatmgrd[4];
  int i, j, k, ih, jh, kh;
  iList *help;
  char outflag;
  double d2;

  /* assign pt to atomgridcell */
  if (0 != pt_to_cell(iptatmgrd, pt, atomgrid)) {
    cerr << "pt outside the grid in pt_inside_atom()" << endl;
    exit(0);
    return 0;
  }

  outflag = 0;
  for (i = -(*atomgrid).look; i <= (*atomgrid).look; i++) {
    ih = iptatmgrd[1] + i;
    if (!((ih > (*atomgrid).imax[1]) || (ih < (*atomgrid).imin[1]))) {
      for (j = -(*atomgrid).look; j <= + (*atomgrid).look; j++) {
        jh = iptatmgrd[2] + j;
        if (!((jh > (*atomgrid).imax[2]) || (jh < (*atomgrid).imin[2]))) {
          for (k = -(*atomgrid).look; k <= + (*atomgrid).look; k++) {
            kh = iptatmgrd[3] + k;
            if (!((kh > (*atomgrid).imax[3]) || (kh < (*atomgrid).imin[3]))) {
              help = &(*atomgrid).root[ih][jh][kh];
              /* check the distance */
              while ((*help).next != NULL) {
                help = (*help).next;
                d2 = get_sqr_dist(pt[1], pt[2], pt[3], xyz[(*help).number][1], xyz[(*help).number][2], xyz[(*help).number][3]);
//                            d2=sqr_distance(pt,xyz[(*help).number],3);
                if (d2 < r2((*help).number)) {
                  return 1; /* inside the atom */
                }
                if (d2 < pr2((*help).number))
                  {outflag = 1;} /* inside the solute */
              }
            }
          }
        }
      }
    }
  }

  if (outflag == 1) return -1;    /* point is between atom and dotsurface */
  else if (outflag == 0) return  0; /* point is in solvent */
  else {
    cerr << "Problem in pt_inside_atom()" << endl;
    exit(0);
    return -100;
  }
}

/** Check if a point is inside a small void

Checks if the point is inside a void if its not inside the atom volume. If the atom is not inside the atom volume, the atom is in a void if the distance between the closest point of the dot surface and the point is larger as the prob spere radius
\param pt The point to evaluate
\param dotgrid dot surface mapped on a grid used as neighbourlist
\param dots Dots of the surface
\param prob2 (atom radius + probe sphere radius)^2
\return int 1 means point is inside the atom volume, -1 point is between atom an probesphere, 0 means point is in solvent

*/
int monte_carlo_run::pt_inside_void(double *pt, iListGrid *dotgrid, double **dots, double prob2) {
  int iptdotgrd[4];
  int i, j, k, ih, jh, kh;
  iList *help;
  double d2;

  /* assign pt to atomgridcell */
  if (0 != pt_to_cell(iptdotgrd, pt, dotgrid)) {
    cerr << "pt outside the grid in pt_inside_atom()" << endl;
    exit(0);
    return 0;
  }

  for (i = -(*dotgrid).look; i <= (*dotgrid).look; i++) {
    ih = iptdotgrd[1] + i;
    if (!((ih > (*dotgrid).imax[1]) || (ih < (*dotgrid).imin[1]))) {
      for (j = -(*dotgrid).look; j <= + (*dotgrid).look; j++) {
        jh = iptdotgrd[2] + j;
        if (!((jh > (*dotgrid).imax[2]) || (jh < (*dotgrid).imin[2]))) {
          for (k = -(*dotgrid).look; k <= + (*dotgrid).look; k++) {
            kh = iptdotgrd[3] + k;
            if (!((kh > (*dotgrid).imax[3]) || (kh < (*dotgrid).imin[3]))) {
              help = &(*dotgrid).root[ih][jh][kh];
              /* check the distance */
              while ((*help).next != NULL) {
                help = (*help).next;
                d2 = get_sqr_dist(pt[1], pt[2], pt[3], dots[(*help).number][1], dots[(*help).number][2], dots[(*help).number][3]);
                //d2=sqr_distance(pt,dots[(*help).number],3);
                if (d2 <= prob2) {
                  return 0; /* in solvent */
                }
              }
            }
          }
        }
      }
    }
  }
  return 1; /* in void */
}
