/* Matthias Ullmann                                     */
/* Bayreuth, September 2006                             */
/*                                                      */
/* .....................................................*/

#include "test-volume.h"
/**Main Program Flow

  After reading the setup, the solvent accessible surface of the whole protein is calcualted by the mit_surf class, followed by an obtional clustering of the dot surface points. Using the calculated dot surface, a Monte Carlo algorithm is applied to calculate the volume of the Protein and defining solvent points, void points and points inside the atom volume. These points are mapped on a grid. Taking all points from this gird which are marked as solvent, cavities (groups of solvent points which have no direct connection to another group of solvent points) are isolated. Obtional the protein surface is searched for open and closed mouth clefts. Two methods are available to search the protein surface for clefts. All remaining solvent points are used to define a membrane of dummy atoms with given z coordinates. The previously defined cavities are afterwards refined by calculating the volume again by the same Monte Carlo algorithm but with a finer grid resolution. Finally all cavities are filled with water molecules and the positions of these water molecules are relaxed by an Monte Carlo algorithm to gain water positions which are as far apart from each other as possible.
*/
int main(int argc, char *argv[]) {
  /* header */
  printf("This is vol version 1.2\n");
  if (argc != 2) {
    fprintf(stderr, "Usage: vol <molname>\n");
    exit(0);
  }
  // Variables we need
  double maxrad; /* Maximum radius of all atoms*/
  double vol; /* Protein Volume */
  int nmc; /* Number of Monte Carlo steps */
  int ncav; /* Number of cavities */
  int flag; /* Loop control*/
  MUS_atom p; /* Data structure holding the protein coordinates and atom information*/
  iListGrid atomgrid;/* Interger grid with atoms mapped to this grid */
  iListGrid dotgrid; /* Integer grid with the dot surface mapped to this grid*/
  iGrid solvgrid; /* Integer Grid holding the information if the point is located in the solvent, void or atom volume */
  iGrid cavgrid; /* Integer grid holding the information for each point belongings to cavities. Point i belongs ot cavity cavgrid[i] */
  long seed; /* Seed for the Monte Carlo algorithms */
  time_t tim; /* Init seed */
  monte_carlo_run mc_run; /* Wrapper for the Monte Carlo algorithm */
  adv_doubly_LL<cavity2> all_cavs; /* Doubly linked list for the cavities, one element is one cavity */
  char pqr_file[100]; /* Finlename char vector*/
  iGrid cav_solvgrid; /* solvgird for the cavity refinement */
  Sys sys(argv[1]); /* Holding the System parameters provided in the setup file*/


  // Reading system Parameters
  sys.read_setup();
  sys.print_param(cout);

  //delete cav folder if present and create new one
  system("rm -rf cav");
  system("mkdir cav");

  flag = 0;

  // Read pqr File
  sprintf(pqr_file, "%s%s", sys.molname.c_str(), ".pqr");
  read_PQR(pqr_file, &p);
  // get min/max coordinates of the protein and save them in mc_run.vars
  min_max_coord(mc_run.vars.min, mc_run.vars.max, p.xyz, p.number_of_atoms);
  // Store user input (setup file) in mc_run.vars
  mc_run.vars.nmc = sys.nmc;
  mc_run.vars.prob = sys.probe;
  mc_run.vars.nprt = 100000;
  nmc    = sys.nmc;

  // Initialize the grids and the seed
  atomgrid.spacing = 4.0;
  dotgrid.spacing  = sys.start_grid;
  solvgrid.spacing = sys.start_grid;
  seed = -time(&tim);
  // start random number generator
  srand(seed); 

  /* Just to prevent somebody from bursting the memory */
  if (max_element(p.radius, p.number_of_atoms) > 5) {
    fprintf(stderr, "\n WARNING: There are atoms with a radius bigger than 5A. Are you sure?\n\n");
  };

  /* calculate the solvent dot surface */
  fprintf(stdout, "================================================\n");
  fprintf(stdout, " SURFACE CALCULATION\n");
  mit_surface mit_surf(p.surface, p.xyz, p.radius, p.number_of_atoms, sys.surf_points, sys.probe, 1);
  cout << "Total Surface: " << mit_surf.get_surface() << endl;

  p.charge = p.surface;
  fprintf(stdout, "================================================\n");

  /* map the protein and dot surface to the grid */
  mc_run.vars.maxrad = maxrad = max_element(p.radius, p.number_of_atoms);
  ProteinToGrid(&atomgrid, &p, maxrad + sys.probe);
  SurfaceToGrid(&dotgrid, mit_surf.get_DotSurf_p());
  ProteinToiGrid(&solvgrid, &p, maxrad + sys.probe, &mc_run.vars);
  if (sys.blab > 1) {
    fprintf(stdout, "================================================\n");
    fprintf(stdout, " GRID INFORMATION\n");
    fprintf(stdout, "\n\n Atom Grid\n");
    iListGrid_info(stdout, &atomgrid);
    fprintf(stdout, "\n\n Dot Surface Grid\n");
    iListGrid_info(stdout, &dotgrid);
    fprintf(stdout, "\n\n Solvent Grid\n");
    iGrid_info(stdout, &solvgrid);
    fprintf(stdout, "================================================\n");
  }
  // If we want to cluster the dot Surface, do this and print the output to files
  if (sys.surface_clusters >= 1) {
    sprintf(pqr_file, "%s_surf%s", sys.molname.c_str(), ".pdb");
    ofstream Surfout(pqr_file);
    mit_surf.print_dotsurf_pqr(Surfout);
    sprintf(pqr_file, "%s_surf%s", sys.molname.c_str(), ".dx");
    mit_surf.cluster(sys.surface_clusters, dotgrid);
    mit_surf.print_clusters_pqr_surf();
    mit_surf.print_clusters_pqr_cav();
    mit_surf.print_clusters_surfaces();
  };
  if(sys.break_after_surface){return 0;}
  // For testing reasons, calculate the volume of the protein taking atoms as spheres
  /* sum the volume over all spheres */
  vol = 0.0;
  for (int i = 1; i <= p.number_of_atoms; i++) {
    vol   = vol + 4.0 / 3.0 * M_PI * p.radius[i] * p.radius[i] * p.radius[i];
  }
  if (sys.blab > 1) {
    fprintf(stdout, "Volume of all spheres: %10.6f A^3\n", vol);
    fprintf(stdout, "================================================\n");
  }

  /* Monte Carlo integration of the volume */

  fprintf(stdout, "================================================\n");
  fprintf(stdout, "VOLUME CALCULATION\n");
  mc_run.vars.blab = 1;
  vol = mc_run.volume_mc(&atomgrid, &dotgrid, &solvgrid, &p, mit_surf.get_DotSurf_p(),&seed);
  smooth_solvgrid(mc_run.vars.min, mc_run.vars.max, mc_run.vars.maxrad, solvgrid);
  fprintf(stdout, "================================================\n");
  mc_run.vars.blab = 0;

  /* Cavity Analysis */
  fprintf(stdout, "================================================\n");
  fprintf(stdout, "CAVITY CALCULATION\n");
  cavgrid.spacing    = solvgrid.spacing;
  cavgrid.inv_spacing = 1.0 / cavgrid.spacing;
  set_iGrid(cavgrid, solvgrid.imin[1], solvgrid.imin[2], solvgrid.imin[3], solvgrid.imax[1], solvgrid.imax[2], solvgrid.imax[3]);

  cavgrid.grid = i3tensor(cavgrid.imin[1], cavgrid.imax[1],
                          cavgrid.imin[2], cavgrid.imax[2],
                          cavgrid.imin[3], cavgrid.imax[3]);
  initialize_iGrid(cavgrid, 0);
  // Graph search algorithm for searching Cavities
  ncav =  search_cav(solvgrid, cavgrid);
  fprintf(stdout, "Number of cavities found: %d\n", ncav);
  if (ncav > 500){printf("This is not good for your file system! We got more than 500 cavities! We'll stop here!\n"); exit(0);};
  int overall_cav_count = 0;
  int nreal_cav = ncav;
  // If we want to search for surface clefts
  if (sys.cleft_method > 0) {
    cout << "Searching for clefts with Method " << sys.cleft_method << endl;
    cout << "CavRel is set to " << sys.nsour_cleft << endl;
    cout << "CleftDim is set to " << sys.cleft_d << endl;
    if (sys.cleft_method == 1) {
      check_for_clefts_sphere(cavgrid, sys.nsour_cleft);
    } else {
      check_for_clefts_first_solv(cavgrid, sys.cleft_d, sys.nsour_cleft);
    };
    // Cluster surface clefts
    ncav =  search_cav(solvgrid, cavgrid, ncav);
    cout << "Number of cavities including clefts: " << ncav << endl;
  }
  // Membrane Implementation
  if (sys.memb_z_max - sys.memb_z_min > 0.1) {
    cavity2 *complete_memb = new cavity2;
    (*complete_memb).init(cavgrid, solvgrid, 1, 10, 1, 1.7, 1);
    char fname[100];
    sprintf(fname, "cav/memb.pqr");
    ofstream tc(fname);
    sprintf(fname, "cav/memb_cutof.pqr");
    ofstream tc_alt(fname);
    sprintf(fname, "cav/memb_rest.pqr");
    ofstream tc_alt2(fname);
    cavity2 *memb = new cavity2;
    // Cut of all points above and below z_max z_min and check the core region
    memb->extract_membrane_slice(*complete_memb, sys.memb_z_min, sys.memb_z_max, sys.dummy_rad, sys.core_z_min, sys.core_z_max, sys.core_dim);
    // From now on, we have the membrane slice in memb, and the rest of all in complete_memeb!
    // Reduce the thickness of the membrane to the given thickness
    memb->reduce_cav_to_layer(p.xyz, p.number_of_atoms, sys.membrane_dim);
    // Reduce the membrane above and below z_min and z_max to the given thickness as well
    complete_memb->reduce_cav_to_layer(p.xyz, p.number_of_atoms, sys.membrane_dim);
    // print the membrane
    memb->print_pqr(tc, 1);
    // print what we have deleted from the membrane to make it somwhat more save
    complete_memb->print_pqr_membrane_slice(tc_alt, tc_alt2, sys.memb_z_min, sys.memb_z_max, 1);

  }
  printf("\nCavity Volume refinement.\n\n");
  for (int cav_count = 2; cav_count <= ncav; cav_count++) {
    if (cav_count <= nreal_cav) {
      cout << "Refining cavity " << cav_count << endl;
      /* Refine Volume calculation of cavities only
        Refineing clefts yields nasty boxes since we can not define a clear border line between the cleft and the solvent
      */
      mc_run.vars.min[1] = cavgrid.imax[1];
      mc_run.vars.min[2] = cavgrid.imax[2];
      mc_run.vars.min[3] = cavgrid.imax[3];
      mc_run.vars.max[1] = cavgrid.imin[1];
      mc_run.vars.max[2] = cavgrid.imin[2];
      mc_run.vars.max[3] = cavgrid.imin[3];
      cout << cavgrid.spacing << endl;
      // Define the box only around the cavity
      for (int i = cavgrid.imin[1]; i <= cavgrid.imax[1]; i++) {
        for (int j = cavgrid.imin[2]; j <= cavgrid.imax[2]; j++) {
          for (int k = cavgrid.imin[3]; k <= cavgrid.imax[3]; k++) {
            if (cavgrid.grid[i][j][k] == cav_count) {
              if (i < mc_run.vars.min[1]){mc_run.vars.min[1] = i;}
              if (i > mc_run.vars.max[1]){mc_run.vars.max[1] = i;}
              if (j < mc_run.vars.min[2]){mc_run.vars.min[2] = j;}
              if (j > mc_run.vars.max[2]){mc_run.vars.max[2] = j;}
              if (k < mc_run.vars.min[3]){mc_run.vars.min[3] = k;}
              if (k > mc_run.vars.max[3]){mc_run.vars.max[3] = k;}
            };
          };
        };
      };

      mc_run.vars.min[1] *=  cavgrid.spacing;
      mc_run.vars.min[2] *=  cavgrid.spacing;
      mc_run.vars.min[3] *=  cavgrid.spacing;
      mc_run.vars.max[1] *=  cavgrid.spacing;
      mc_run.vars.max[2] *=  cavgrid.spacing;
      mc_run.vars.max[3] *=  cavgrid.spacing;
      double tmp_boxvol = get_box_volume(solvgrid);
      int nmc = (int) ceil((mc_run.vars.nmc * tmp_boxvol) / solvgrid.spacing);

      mc_run.vars.maxrad = 1.0; /* Dirty Hack?? */
      cav_solvgrid.spacing = sys.cav_grid;
      ProteinToiGrid(&cav_solvgrid, &p, mc_run.vars.maxrad, &mc_run.vars);
      vol = mc_run.volume_mc(&atomgrid, &dotgrid, &cav_solvgrid, &p, mit_surf.get_DotSurf_p(),&seed);

      iGrid cavgrid_local;
      cavgrid_local.spacing    = cav_solvgrid.spacing;
      cavgrid_local.inv_spacing = 1.0 / cavgrid_local.spacing;
      set_iGrid(cavgrid_local, cav_solvgrid.imin[1], cav_solvgrid.imin[2], cav_solvgrid.imin[3], cav_solvgrid.imax[1], cav_solvgrid.imax[2], cav_solvgrid.imax[3]);
      cavgrid_local.grid = i3tensor(cavgrid_local.imin[1], cavgrid_local.imax[1], cavgrid_local.imin[2], cavgrid_local.imax[2], cavgrid_local.imin[3], cavgrid_local.imax[3]);
      initialize_iGrid(cavgrid_local, 0);
      int ncav_local = search_cav(cav_solvgrid, cavgrid_local);
      // Now we have to cluster the result of the refinement again. There can be artefacts from other cavities in the same box, one cavity with only a small bridge between two larger regions could have been seperated...
      for (int cav_count_local = 1; cav_count_local <= ncav_local; cav_count_local++) {
        cavity2 *local_c = new cavity2;
        /* calculate Volume of Cavities */
        /* First of all define a new iListgrid and a new iGrid, containing the cavity */
        tmp_boxvol = get_box_volume(cav_solvgrid);
        nmc = (int) ceil((mc_run.vars.nmc * tmp_boxvol) / cav_solvgrid.spacing);
        overall_cav_count++;
        local_c->init(cavgrid_local, cav_solvgrid, cav_count_local, nmc, overall_cav_count, sys.dummy_rad, 1);
        if (local_c->get_volume() >= sys.min_vol) {
          cout << "Cavity added: "<< fixed << setprecision(2) << local_c->get_volume() << " as cavity number " << local_c->get_running_num() << endl;
          all_cavs.add_tail(local_c);
        } else {
            cout << "Cavity (" << local_c->get_volume() << " A^3) smaller than min_mol (" << sys.min_vol << " A^3). Discarded" << endl;
          delete local_c;
        }
      }
      // Free some of the memory
      free_i3tensor(cav_solvgrid.grid, cav_solvgrid.imin[1], cav_solvgrid.imax[1], cav_solvgrid.imin[2], cav_solvgrid.imax[2], cav_solvgrid.imin[3], cav_solvgrid.imax[3]);
    } else {
      cout << "Storing cleft " << cav_count << endl;
      cavity2 *local_c = new cavity2;
      // For clefts, we only take what we got from the first calcualtion, no refinement here
      double tmp_boxvol = get_box_volume(solvgrid);
      nmc = (int) ceil((mc_run.vars.nmc * tmp_boxvol) / solvgrid.spacing);
      overall_cav_count++;

      local_c->init(cavgrid, solvgrid, cav_count, nmc, overall_cav_count, sys.dummy_rad, 2);
      if (local_c->get_volume() >= sys.min_vol) {
        cout << "Cleft added: " << local_c->get_volume() << endl;
        all_cavs.add_tail(local_c);
      } else {
        cout << "Cleft (" << local_c->get_volume() << " A^3) smaller than min_mol(" << sys.min_vol << "A^3). Discarded" << endl;
        delete local_c;
      }
    }
  }
  // Free some memory
  free_i3tensor(cavgrid.grid, cavgrid.imin[1], cavgrid.imax[1], cavgrid.imin[2], cavgrid.imax[2], cavgrid.imin[3], cavgrid.imax[3]);
  adv_doubly_LL<cavity2>::iterator iter = all_cavs.begin();

  // Now we have to check, if we have added artifically cavities twice or fragments of cavities twice.

  while (iter != NULL) {
    flag = 0;
    // first of all, check if a cavity contains points at the borderlines of its surrounding box. If yes, this cavity is most probably an artifact of the refinement, since the box for the refinement is choosen largen than the cavity should be. Cavities with points at the box border are artifacts from cavities in the same area as the one we wanted to refine, but belonging to another cavity
    if ((*iter)->get_kind() == 1) {
      if ((*iter)->check_border_points()) {
        cout << "Cavity " << (*iter)->get_running_num() << " don't touches border." << endl;
        flag = 1;
      } else {
        cout << "Cavity " << (*iter)->get_running_num() << " touches border so this might be artificial." << endl;
        cout << "Deleting this cavity!" << endl;
        all_cavs.delete_element(iter);
      }
    } else{flag = 1;};
    if (flag)iter++;
  }

  if (sys.delete_overlapping) {
    // Same purpose as before. If we have found some points in two different cavities, we will merge the cavities
    cout << "Check cavities for duplicates" << endl;
    iter = all_cavs.begin();
    while (iter != NULL) {
      if ((*iter)->get_kind() == 1) {
        cout << "Checking cavity " << (*iter)->get_running_num() << " with " << (*iter)->get_volume() << "A^3" << endl;
      } else {
        cout << "Checking cleft " << (*iter)->get_running_num() << " with " << (*iter)->get_volume() << "A^3" << endl;
      }
      adv_doubly_LL<cavity2>::iterator iter2 = all_cavs.begin();
      while (iter2 != NULL) {
        //cout << "and cavity " << (*iter2)->get_running_num() << endl;
        flag = 0;
        if (iter != iter2 && (*iter2)->get_running_num() > (*iter)->get_running_num()) {
          if ((*iter2)->check_duplicate(**iter)) {
            cout << "    dublicates " << (*iter)->get_running_num() << " " << (*iter2)->get_running_num() << endl;
            if ((*iter)->get_kind() == (*iter2)->get_kind()) {
              cout << "Merging..." << endl;
              (*iter2)->merge_with(**iter);
              all_cavs.delete_element(iter2);
              flag = 2;
            } else {
              if ((*iter)->get_volume() < (*iter2)->get_volume()) {
                cout << "      Deleting smaller one: " << (*iter)->get_running_num() << endl;
                all_cavs.delete_element(iter);
                flag = 1;
              } else {
                cout << "      Deleting smaller one: " << (*iter2)->get_running_num() << endl;
                flag = 2;
                all_cavs.delete_element(iter2);

              }
            }
          }
        }
        if (flag != 2){iter2++;}
      }
      if (flag != 1){iter++;}
    }
  } else {
    cout << "Not checking for duplicate cavities!" << endl;
  }
  // Print the pqr files of all cavities and clefts.
  char fname[100];
  iter = all_cavs.begin();
  while (iter != NULL) {

    if ((*iter)->get_kind() == 1) {
      sprintf(fname, "cav/cav_%d.pqr", (*iter)->get_running_num());
    } else {
      sprintf(fname, "cav/real_cleft_%d.pqr", (*iter)->get_running_num());
    }
    ofstream out(fname);
    cout << "printing cav " << (*iter)->get_running_num() << " to " << fname << endl;
    (*iter)->print_pqr(out, (*iter)->get_running_num());
    out.close();
    iter++;
  }


  // Finally place water molecules in all cavities and clefts which are large enough
  cout << "Placing water molecules" << endl;
  iter = all_cavs.begin();
  int nwater = 1;
  ofstream w_out("cav/water.pdb");
  while (iter != NULL) {
    cout << (*iter)->add_water_molecules2(sys.water_vol) << " waters placed in cavity " << (*iter)->get_running_num() << endl;
    nwater = (*iter)->print_waters(w_out, nwater);
    iter++;
  }
  cout << nwater << " Water molecules placed in total. " << endl;
  free_iList_3tensor(dotgrid.root, dotgrid.imin[1], dotgrid.imax[1], dotgrid.imin[2], dotgrid.imax[2], dotgrid.imin[3], dotgrid.imax[3]);
  cout << "Everything is done! By!" << endl;
  return 0;

}
